from collections import OrderedDict as od
from os.path import join, exists, basename, splitext
from pyaerocom import const
from pyaerocom.tstype import TsType
from pyaerocom.exceptions import FileConventionError
from configparser import ConfigParser

class FileConventionRead(object):
    """Class that represents a file naming convention for reading Aerocom files

    Attributes
    ----------
    name : str
        name of this convention (e.g. "aerocom3")
    file_sep : str
        filename delimiter for accessing different variables
    year_pos : int
        position of year information in filename after splitting using
        delimiter :attr:`file_sep`
    var_pos : int
        position of variable information in filename after splitting using
        delimiter :attr:`file_sep`
    ts_pos : int
        position of information of temporal resolution in filename after
        splitting using delimiter :attr:`file_sep`
    vert_pos : int
        position of information about vertical resolution of data
    data_id_pos : int
        position of data ID
    """
    _io_opts = const
    AEROCOM3_VERT_INFO = {'2d'  : ['surface', 'column', 'modellevel'],
                          '3d'  : ['modellevelatstations']}

    def __init__(self, name="aerocom3", file_sep="_", year_pos=None,
                 var_pos=None, ts_pos=None, vert_pos=None, data_id_pos=None,
                 from_file=None):

        self.name = name
        self.file_sep = file_sep

        self.year_pos = year_pos
        self.var_pos = var_pos
        self.ts_pos = ts_pos
        self.vert_pos = vert_pos
        self.data_id_pos = data_id_pos

        if from_file is not None:
            self.from_file(from_file)
        else:
            try:
                self.import_default(self.name)
            except Exception:
                pass

    @property
    def info_init(self):
        """Empty dictionary containing init values of infos to be
        extracted from filenames
        """
        return od(year=None, var_name=None, ts_type=None, vert_code='',
                  is_at_stations=False, data_id='')

    def from_file(self, file):
        """Identify convention from a file

        Currently only two conventions (aerocom2 and aerocom3) exist that are
        identified by the delimiter used.

        Parameters
        ----------
        file : str
            file path or file name

        Returns
        -------
        FileConventionRead
            this object (with updated convention)

        Raises
        ------
        FileConventionError
            if convention cannot be identified

        Example
        -------
        >>> from pyaerocom.io import FileConventionRead
        >>> filename = 'aerocom3_CAM5.3-Oslo_AP3-CTRL2016-PD_od550aer_Column_2010_monthly.nc'
        >>> print(FileConventionRead().from_file(filename))
        pyaeorocom FileConventionRead
        name: aerocom3
        file_sep: _
        year_pos: -2
        var_pos: -4
        ts_pos: -1
        """

        if basename(file).count("_") >= 4:
            self.import_default("aerocom3")
        elif basename(file).count(".") >= 4:
            self.import_default("aerocom2")
        else:
            raise FileConventionError('Could not identify convention from '
                                      'input file {}'.format(basename(file)))
        self.check_validity(file)
        return self

    def check_validity(self, file):
        """Check if filename is valid"""
        info = self.get_info_from_file(file)
        year = info["year"]
        if not TsType.valid(info['ts_type']):
            raise FileConventionError("Invalid ts_type %s in filename %s"
                                           %(info['ts_type'], basename(file)))
        elif not (const.MIN_YEAR <= year <= const.MAX_YEAR):
            raise FileConventionError("Invalid year %d in filename %s"
                          %(info['year'], basename(file)))

    def _info_from_aerocom3(self, file):
        """Extract info from filename Aerocom 3 convention

        Parameters
        -----------
        file : str
            netcdf file name

        Returns
        -------
        dict
            dictionary containing infos that were extracted from filename
        """
        # init result dictionary
        info = self.info_init

        spl = splitext(basename(file))[0].split(self.file_sep)
        # phase 3 file naming convention
        try:
            info["year"] = int(spl[self.year_pos])
        except Exception:
            msg = ("Failed to extract year information from file {} "
                   "using file convention Aerocom 3".format(basename(file),
                                                            self.name))
            raise FileConventionError(msg)
        try:
            # include vars for the surface
            if spl[self.vert_pos].lower() in self.AEROCOM3_VERT_INFO['2d']:
                info["var_name"] = spl[self.var_pos]
            # also include 3d vars that provide station based data
            # and contain the string vmr in this case the variable name has to
            # be slightly changed to the  aerocom phase 2 naming
            elif spl[self.vert_pos].lower() in self.AEROCOM3_VERT_INFO['3d']:
                if 'vmr' in spl[self.var_pos]:
                    info['var_name'] = spl[self.var_pos].replace('vmr', 'vmr3d')
                else:
                    info['var_name'] = spl[self.var_pos]
            else:
                raise FileConventionError('Invalid file name (Aerocom 3 '
                                          'naming convention).\n'
                                          '{}\nInvalid string identifier for '
                                          'vertical coordinate: {}'
                                          .format(file, spl[self.vert_pos]))
        except Exception as e:
            raise FileConventionError('Failed to extract variable name from '
                                      'file {} using file convention {}.\n'
                                      'Error: {}'.format(basename(file),
                                                         self.name, repr(e)))
        try:
            info["ts_type"] = spl[self.ts_pos]
        except Exception:
            raise FileConventionError('Failed to extract ts_type from '
                                      'file {} using file convention {}'
                                      .format(basename(file), self.name))
        try:
            info["vert_code"] = spl[self.vert_pos]
        except Exception:
            raise FileConventionError('Failed to extract vert_code from '
                                      'file {} using file convention {}'
                                      .format(basename(file), self.name))

        try:
            info["data_id"] = self.file_sep.join(spl[self.data_id_pos:self.var_pos])
        except Exception:
            raise FileConventionError('Failed to extract model name from '
                                      'file {} using file convention {}'
                                      .format(basename(file), self.name))
        if'atstations' in file.lower():
            info['is_at_stations'] = True
        return info

    def _info_from_aerocom2(self, file):
        """Extract info from filename Aerocom 2 convention

        Parameters
        -----------
        file : str
            netcdf file name

        Returns
        -------
        dict
            dictionary containing infos that were extracted from filename
        """
        info = self.info_init
        if self.file_sep == ".":
            spl = basename(file).split(self.file_sep)
        else:
            spl = splitext(basename(file))[0].split(self.file_sep)
        try:
            info["year"] = int(spl[self.year_pos])
        except Exception:
            raise FileConventionError('Failed to extract year information '
                                      'from file {} using file '
                                      'convention {}'
                                      .format(basename(file), self.name))
        try:
            info["var_name"] = spl[self.var_pos]
        except Exception:
            raise FileConventionError('Failed to extract variable information '
                                      'from file {} using file '
                                      'convention {}'
                                      .format(basename(file), self.name))
        try:
            info["ts_type"] = spl[self.ts_pos]
        except Exception:
            raise FileConventionError('Failed to extract ts_type '
                                      'from file {} using file '
                                      'convention {}'
                                      .format(basename(file), self.name))

        try:
            info['data_id'] = '.'.join(spl[self.data_id_pos:self.ts_pos])
        except Exception:
            raise FileConventionError('Failed to extract name '
                                      'from file {} using file '
                                      'convention {}'
                                      .format(basename(file), self.name))
        if 'atstations' in file.lower():
            raise Exception('Developers: please debug (file convention '
                            'Aerocom 2 should not have atstations '
                            'encoded in file name)')
        return info

    def get_info_from_file(self, file):
        """Identify convention from a file

        Currently only two conventions (aerocom2 and aerocom3) exist that are
        identified by the delimiter used.

        Parameters
        ----------
        file : str
            file path or file name

        Returns
        -------
        OrderedDict
            dictionary containing keys `year, var_name, ts_type` and
            corresponding variables, extracted from the filename

        Raises
        ------
        FileConventionError
            if convention cannot be identified

        Example
        -------
        >>> from pyaerocom.io import FileConventionRead
        >>> filename = 'aerocom3_CAM5.3-Oslo_AP3-CTRL2016-PD_od550aer_Column_2010_monthly.nc'
        >>> conv = FileConventionRead("aerocom3")
        >>> info = conv.get_info_from_file(filename)
        >>> for item in info.items(): print(item)
        ('year', 2010)
        ('var_name', 'od550aer')
        ('ts_type', 'monthly')
        """
        if self.name == 'aerocom3':
            return self._info_from_aerocom3(file)
        elif self.name == 'aerocom2':
            return self._info_from_aerocom2(file)

    def string_mask(self, data_id, var, year, ts_type, vert_which=None):
        """Returns mask that can be used to identify files of this convention

        Parameters
        ----------
        data_id : str
            experiment ID (e.g. GISS-MATRIX.A2.CTRL)
        var : str
            variable string ID (e.g. "od550aer")
        year : int
            desired year of observation (e.g. 2012)
        ts_type : str
            string specifying temporal resolution (e.g. "daily")

        Example
        -------

        import re
        conf_aero2 = FileConventionRead(name="aerocom2")
        conf_aero3 = FileConventionRead(name="aerocom2")

        var = od550aer
        year = 2012
        ts_type = "daily"

        match_str_aero2 = conf_aero2.string_mask(var, year, ts_type)

        match_str_aero3 = conf_aero3.string_mask(var, year, ts_type)

        """
        if ts_type is None:
            ts_type = '*'
        if self.name == "aerocom2":
            if vert_which is not None:
                raise FileConventionError('Specification of vert_which ({}) is '
                                          'not supported for '
                                          'aerocom2 naming convention'
                                          .format(vert_which))

            return ".".join(['.*', data_id, ts_type, var, str(year), 'nc'])
        elif self.name == "aerocom3":
            if vert_which is None:
                vert_which = '.*'
            return "_".join(['.*',  data_id, var, vert_which, str(year), ts_type]) + '.nc'
        else:
            raise NotImplementedError("File matching mask for convention %s "
                                      "not yet defined..." %self.name)

    def import_default(self, name):
        """Checks and load default information from database"""
        from pyaerocom import __dir__
        fpath = join(__dir__, "data", "file_conventions.ini")
        if not exists(fpath):
            raise IOError("File conventions ini file could not be found: %s"
                          %fpath)
        conf_reader = ConfigParser()
        conf_reader.read(fpath)
        if not name in conf_reader:
            raise NameError("No default available for %s" %name)
        self.name = name
        for key, val in conf_reader[name].items():
            if key in self.__dict__:
                try:
                    val = int(val)
                except Exception:
                    pass
                self.__dict__[key] = val

    def from_dict(self, new_vals):
        """Load info from dictionary

        Parameters
        ----------
        new_vals : dict
            dictionary containing information

        Returns
        -------
        self
        """
        for k, v in new_vals.items():
            if k in self.__dict__:
                self.__dict__[k] = v
        return self

    def to_dict(self):
        """Convert this object to ordered dictionary"""
        return od(name = self.name,
                  file_sep = self.file_sep,
                  year_pos = self.year_pos,
                  var_pos = self.var_pos,
                  ts_pos = self.ts_pos,
                  vert_pos = self.vert_pos,
                  data_id_pos = self.data_id_pos)

    def __repr__(self):
       return ("%s %s" %(self.name, super(FileConventionRead, self).__repr__()))

    def __str__(self):
        s = "\npyaeorocom FileConventionRead"
        for k, v in self.to_dict().items():
            s += "\n%s: %s" %(k, v)
        return s

if __name__=="__main__":
    conf = FileConventionRead()

    print(conf)

    d = od(name = "Fake",
           file_sep = 10,
           year_pos = -6,
           var_pos = 15,
           ts_pos = 3)
    print(conf.from_dict(d))
    try:
        conf.import_default("blaaa")
    except NameError:
        print("Works as expected")
    conf.import_default("aerocom3")
    print(conf)

    conf = FileConventionRead(name="aerocom2")
    print(conf)

    fname = 'aerocom3_TM5_AP3-INSITU_vmrch4_ModelLevelAtStations_2010_monthly.nc'

    print('\nFrom file: {}'.format(conf.from_file(fname)))
    print(conf.get_info_from_file(fname))

    ff = FileConventionRead(from_file='aerocom.CALIOP3.monthly.ec5323Ddust.2006.nc')
    print(ff)



import warnings
def filter_warnings(apply, categories=None, messages=None):
    """
    Decorator that can be used to filter particular warnings

    Parameters
    ----------
    apply : bool
        if True warnings will be filtered, else not.
    categories : list, optional
        list of warning categories to be filtered. E.g.
        [UserWarning, DeprecationWarning]. The default is None. For each
        `<entry>` :func:`warnigns.filterwarnings('ignore', category=<entry>)`
        is called.
    messages : list, optional
        list of warning messages to be filtered. E.g.
        ['Warning that can safely be ignored']. The default is None. For each
        `<entry>` :func:`warnigns.filterwarnings('ignore', message=<entry>)`
        is called.

    Example
    -------
    @filter_warnings(categories=[UserWarning, DeprecationWarning],
                     messages=['I REALLY'])
    def warn_randomly_and_add_numbers(num1, num2):
        warnings.warn(UserWarning('Harmless user warning'))
        warnings.warn(DeprecationWarning('This function is deprecated'))
        warnings.warn(Warning('I REALLY NEED TO REACH YOU'))
        return num1+num2

    """
    if categories is None:
        categories = []
    elif not isinstance(categories, list):
        raise ValueError('categories must be list or None')
    if messages is None:
        messages = []
    elif not isinstance(messages, list):
        raise ValueError('messages must be list or None')
    def decorator(func):
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                if apply:
                    for cat in categories:
                        warnings.filterwarnings('ignore', category=cat)
                    for msg in messages:
                        warnings.filterwarnings('ignore', message=msg)
                return func(*args, **kwargs)
        return wrapper
    return decorator

if __name__ == '__main__':

    @filter_warnings(True,[UserWarning], messages=['Deprecated'])
    def add_num_with_warnings(num1, num2):
        warnings.warn(UserWarning('User Warning'))
        warnings.warn(DeprecationWarning('Deprecated'))
        warnings.warn(Warning('General warning'))
        return num1 + num2

    print(add_num_with_warnings(1, 2))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Helper methods for computation of trends

Note
----
Most methods here are private and not to be used directly. Please use
:class:`TrendsEngine` instead.
"""
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

SEASONS = {'spring'     : [3,4,5],
           'summer'     : [6,7,8],
           'autumn'     : [9,10,11],
           'winter'     : [12, 1, 2]}

SEASON_CODES = {'spring'     : 'MAM',
                'summer'     : 'JJA',
                'autumn'     : 'SON',
                'winter'     : 'DJF'}

MONTHS_CODES = {
                'MAM': 'spring',
                'JJA': 'summer',
                'SON': 'autumn',
                'DJF': 'winter',
                'all': 'all',
                }

def _init_trends_result_dict(start_yr):
    keys = ['pval', 'm', 'm_err',
            'n', 'y_mean', 'y_min', 'y_max',
            'slp', 'slp_err', 'reg0', # data specific
            'slp_{}'.format(start_yr), # period specific
            'slp_{}_err'.format(start_yr), # period specific
            'reg0_{}'.format(start_yr), # period specific
            'data'
            ]
    return dict.fromkeys(keys)

def _compute_trend_error(m, m_err, v0, v0_err):
    """Computes error of trend estimate using gaussian error propagation

    The (normalised) trend is computed as T = m / v0

    where m denotes the slope of a regression line and v0 denotes the
    normalistation value. This method computes the uncertainty of T (delta_T)
    using gaussian error propagation of uncertainties accompanying m and v0.

    Parameters
    ----------
    m : float
        slope in units of <U> yr-1 (where <U> denotes the unit of the data).
        (m -> "montant").
    m_err : float
        slope error (same unit as `m`)
    v0 : float
        normalisation value in units of <U>
    v0_err : float
        error of `v0` (same units as `v0`)

    Returns
    -------
    float
        error of T in computed using gaussian error propagation of trend
        formula in units of %/yr
    """

    delta_sl = m_err / v0
    delta_ref = m * v0_err / v0**2
    return np.sqrt(delta_sl**2 + delta_ref**2) * 100

def _get_season(mon):

    for seas, months in SEASONS.items():
        if mon in months:
            return seas

def _get_unique_seasons(idx):
    seasons = []
    mons = idx.month
    for mon in mons:
        seas = _get_season(mon)
        if not seas in seasons:
            seasons.append(seas)
    return seasons

def _get_season_from_months(months: str) -> str:
    if months in MONTHS_CODES.keys():
        return MONTHS_CODES[months]
    else:
        raise ValueError(f"{months} is not a valid season")


def _mid_season(seas, yr):
    if seas=='spring':
        return np.datetime64('{}-04-15'.format(yr))
    if seas=='summer':
        return np.datetime64('{}-07-15'.format(yr))
    if seas=='autumn':
        return np.datetime64('{}-10-15'.format(yr))
    if seas=='winter':
        return np.datetime64('{}-01-15'.format(yr))
    if seas=='all':
        return np.datetime64('{}-06-15'.format(yr))
    raise ValueError('Invalid input for season (seas):', seas)

def _start_season(seas, yr):
    if seas=='spring':
        return '{}-03-01'.format(yr)
    if seas=='summer':
        return '{}-06-01'.format(yr)
    if seas=='autumn':
        return '{}-09-01'.format(yr)
    if seas=='winter':
        return '{}-12-01'.format(yr-1)
    if seas=='all':
        return '{}-01-01'.format(yr)
    raise ValueError('Invalid input for season (seas):', seas)

def _end_season(seas, yr):
    if seas=='spring':
        return '{}-06-01'.format(yr)
    if seas=='summer':
        return '{}-09-01'.format(yr)
    if seas=='autumn':
        return '{}-12-01'.format(yr)
    if seas=='winter':
        return '{}-03-01'.format(yr)
    if seas=='all':
        return '{}-01-01'.format(yr)
    raise ValueError('Invalid input for season (seas):', seas)


def _find_area(lat, lon, regions_dict=None):
    """Find area corresponding to input lat/lon coordinate

    Parameters
    ----------
    lat : float
        latitude
    lon : float
        longitude

    Returns
    -------
    str
        name of region
    """
    from pyaerocom.region import find_closest_region_coord
    reg = find_closest_region_coord(lat, lon)
    if regions_dict is not None and reg in regions_dict:
        return regions_dict[reg]
    return reg

def _years_from_periodstr(period):
    """Convert period str to start / stop years

    Parameters
    ----------
    period : str
        period str, e.g. '1990-2010'

    Returns
    -------
    int
        start year
    int
        stop year
    """
    return [int(x) for x in period.split('-')]

def _start_stop_period(period):
    """Convert period str to start / stop dates

    Parameters
    ----------
    period : str
        period str, e.g. '1990-2010'

    Returns
    -------
    date
        start datetime
    date
        stop datetime
    """
    from datetime import date
    y0, y1, = _years_from_periodstr(period)
    return (date(y0, 1, 1), date(y1, 12, 31))

def _seas_slice(yr, season):
    pass


def _get_yearly(data, seas, start_yr):
    dates = []
    values = []
    yrs = np.unique(data.index.year)
    for yr in yrs:
        if yr < start_yr: #winter
            continue

        if seas == 'all': #yearly trends
            subset = data.loc[str(yr)]

        else:
            start = _start_season(seas, yr)
            stop = _end_season(seas, yr)
            subset = data.loc[start:stop]

        if len(subset) == 0:
            val = np.nan
        elif np.isnan(subset.values).all():
            val = np.nan

        elif seas=='all':
            seasons = _get_unique_seasons(subset.index)
            if len(seasons) == 4:
                val = np.nanmean(subset.values)
            else:
                val = np.nan
        else:
            val = np.nanmean(subset.values)

        dates.append(_mid_season(seas, yr))
        values.append(val)

    return pd.Series(values, index=dates)

def _init_period_dates(start_year, stop_year, season):
    start_date = _mid_season(season, start_year)
    stop_date = _mid_season(season, stop_year)

    # datetime index covering whole input period (data may not be fully covered
    # in the whole period)
    period_index = pd.date_range(start=start_date,
                                 end=stop_date,
                                 freq=pd.DateOffset(years=1))

    num_dates_period = period_index.values.astype('datetime64[Y]').astype(np.float64)
    return (start_date, stop_date, period_index, num_dates_period)


if __name__ == '__main__':
    import pyaerocom as pya
    plt.close('all')

    r = pya.io.ReadUngridded()
    data = r.read('AeronetSunV3Lev2.daily', vars_to_retrieve='od550aer',
                  file_pattern='Solar*')

    stat = data.to_station_data('Solar*')

    stat.compute_trend('od550aer', 2002, 2012)
    tr = stat.trends['od550aer']

    ax = tr.plot(season='all', period='1995-2017')

    ax = tr.plot(season='all', period='2002-2012')
    plt.show()



"""
Test script for creating variables.ini from HTAP2 excel table
"""
import openpyxl
import string
from collections import OrderedDict as od
from configparser import ConfigParser
import os
tab = 'HTAP2_variables.xlsx'

config = 'variables.ini'

IGNORE = ['freq', 'priority']
if __name__=='__main__':

    book = openpyxl.load_workbook(tab)

    #read sheet Surface
    sheets = ['Surface', 'Column', 'ModelLevel',
              'SurfAtStations (Aerosol)', 'SurfAtStations (Gas)',
              'ModelLevelAtStations']

    cols = ['var_name', 'description', 'standard_name', 'var_type',
                'unit', 'minimum', 'maximum', 'dimensions', 'freq',
                'priority', 'comments_and_purpose']

    result = od()
    multiple = {}
    for sheet_name in sheets:
        sheet = book[sheet_name]
        col_names = string.ascii_uppercase[:len(cols)]

        CTRL_COL = 'C'

        for i, item in enumerate(sheet['A']):

            if sheet[CTRL_COL][i].value is None:
                continue

            var_spec = od()
            var_name = item.value
            if var_name.startswith('HTAP'):
                continue
            if var_name in result:
                try:
                    multiple[var_name] += 1
                except KeyError:
                    multiple[var_name] = 1
                continue
            for j, col in enumerate(col_names):
                val = sheet[col][i].value
                var_spec[cols[j]] = val

            result[var_name] = var_spec

    #df = pd.DataFrame(list(result.values()), columns=cols)
    #df.to_csv(config)

    if not os.path.exists(config):
        open(config, 'a')

    cfg = ConfigParser()
    cfg.read(config)

    errs = {}
    for var, info in result.items():
        if not var in cfg:
            cfg.add_section(var)
        for k, v in info.items():
            if k in IGNORE:
                continue
            try:
                cfg.set(var, k, str(v))
            except ValueError as e:
                errs[var] = '{}: {}'.format(k, repr(e))
    with open(config, 'w') as f:
        cfg.write(f)

    for var, err in errs.items():
        print('{}: {}'.format(var, err))



import abc

from pyaerocom.colocation_auto import Colocator
from pyaerocom._lowlevel_helpers import TypeValidator
from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval.experiment_output import ExperimentOutput


class HasConfig:
    """
    Base class that ensures that evaluation configuration is available

    Attributes
    ----------
    cfg : EvalSetup
        AeroVal experiment setup
    exp_output : ExperimentOutput
        Manages output for an AeroVal experiment (e.g. path locations).

    """
    cfg = TypeValidator(EvalSetup)
    exp_output = TypeValidator(ExperimentOutput)
    def __init__(self, cfg : EvalSetup):
        self.cfg = cfg
        self.exp_output = ExperimentOutput(cfg)

    @property
    def raise_exceptions(self):
        return self.cfg.colocation_opts.raise_exceptions

    @property
    def reanalyse_existing(self):
        return self.cfg.colocation_opts.reanalyse_existing


class ProcessingEngine(HasConfig, abc.ABC):
    """
    Abstract base for classes supposed to do one or more processing tasks

    Requirement for a processing class is to inherit attrs from
    :class:`HasConfig` and, in addition to that, to have implemented a method
    :fun:`run` which is running the corresponding processing task and storing
    all the associated output files, that are read by the frontend.

    One example of an implementation is the
    :class:`pyaerocom.aeroval.modelmaps_engine.ModelMapsEngine`.
    """

    @abc.abstractmethod
    def run(self, *args, **kwargs) -> list:
        """
        Method that runs the processing based on settings in :attr:`cfg`

        Parameters
        ----------
        *args
            positional arguments.
        **kwargs
            Keyword arguments.

        Returns
        -------
        list
            list of output file paths generated by the engine.

        """
        pass


class HasColocator(HasConfig):
    """
    Config class that also has the ability to co-locate
    """
    def _get_diurnal_only(self, obs_name):
        """
        Check if colocated data is flagged for only diurnal processing

        Parameters
        ----------
        obs_name : string
            Name of observational subset
        colocated_data : ColocatedData
            A ColocatedData object that will be checked for suitability of
            diurnal processing.

        Returns
        -------
        diurnal_only : bool
        """
        entry = self.cfg.get_obs_entry(obs_name)
        try:
            diurnal_only = entry['diurnal_only']
        except KeyError:
            diurnal_only = False
        return diurnal_only

    def get_colocator(self, model_name:str=None,
                      obs_name:str=None) -> Colocator:
        """
        Instantiate colocation engine

        Parameters
        ----------
        model_name : str, optional
            name of model. The default is None.
        obs_name : str, optional
            name of obs. The default is None.

        Returns
        -------
        Colocator

        """
        col = Colocator(**self.cfg.colocation_opts)
        if obs_name:
            obs_cfg = self.cfg.get_obs_entry(obs_name)
            col.import_from(obs_cfg)
            col.add_glob_meta(diurnal_only=self._get_diurnal_only(obs_name))
        if model_name:
            mod_cfg = self.cfg.get_model_entry(model_name)
            col.import_from(mod_cfg)
        outdir = self.cfg.path_manager.get_coldata_dir()
        col.basedir_coldata = outdir
        return col


class DataImporter(HasColocator):
    """
    Class that supports reading of model and obs data based on an eval config.

    Depending on a :class:`EvalSetup`, reading of model and obs data may have
    certain constraints (e.g. freq, years, alias variable names, etc.), which
    are / can be specified flexibly for each model and obs entry in an
    analysis setup (:class:`EvalSetup`). Proper handling of these reading
    constraints and data import settings are handled in the
    :class:`pyaerocom.colocation_auto.Colocator` engine, therefore the reading
    in this class is done via the :class:`Colocator` engine.


    """
    def read_model_data(self, model_name, var_name):
        """
        Import model data

        Parameters
        ----------
        model_name : str
            Name of model in :attr:`cfg`,
        var_name : str
            Name of variable to be read.

        Returns
        -------
        data : GriddedData
            loaded model data.

        """
        col = self.get_colocator(model_name=model_name)
        data = col.get_model_data(var_name)

        return data

    def read_ungridded_obsdata(self, obs_name, var_name):
        """
        Import ungridded observation data

        Parameters
        ----------
        obs_name : str
            Name of observation network in :attr:`cfg`
        var_name : str
            Name of variable to be read.

        Returns
        -------
        data : UngriddedData
            loaded obs data.

        """

        col = self.get_colocator(obs_name=obs_name)

        data = col._read_ungridded(var_name)
        return data


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains functionality related to regions in pyaerocom
"""
import numpy as np

from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.helpers_landsea_masks import load_region_mask_xr
from pyaerocom.region_defs import (REGION_DEFS, # all region definitions
                                   HTAP_REGIONS, # list of HTAP regions
                                   OLD_AEROCOM_REGIONS,
                                   REGION_NAMES) # custom names (dict)

class Region(BrowseDict):
    """Class specifying a region

    Attributes
    ----------
    region_id : str
        ID of region (e.g. EUROPE)
    name : str
        name of region (e.g. Europe) used e.g. in plotting.
    lon_range : list
        longitude range (min, max) covered by region
    lat_range : list
        latitude range (min, max) covered by region
    lon_range_plot : list
        longitude range (min, max) used for plotting region.
    lat_range_plot : list
        latitude range (min, max) used for plotting region.
    lon_ticks : list
        list of longitude ticks used for plotting
    lat_ticks : list
        list of latitude ticks used for plotting

    Parameters
    ----------
    region_id : str
        ID of region (e.g. "EUROPE"). If the input region ID is registered as
        a default region in :mod:`pyaerocom.region_defs`, then the default
        information is automatically imported on class instantiation.
    **kwargs
        additional class attributes (see above for available default attributes).
        Note, any attr. values provided by kwargs are preferred over
        potentially defined default attrs. that are imported automatically.
    """
    def __init__(self, region_id=None,**kwargs):

        if region_id is None:
            region_id = 'WORLD'

        if region_id in REGION_NAMES:
            name = REGION_NAMES[region_id]
        else:
            name = region_id

        self.region_id = region_id
        self.name = name

        self.lon_range = None
        self.lat_range = None

        # longitude / latitude range of data in plots
        self.lon_range_plot = None
        self.lat_range_plot = None

        self.lon_ticks = None
        self.lat_ticks = None

        self._mask_data = None
        if region_id in REGION_DEFS:
            self.import_default(region_id)

        self.update(**kwargs)

    def is_htap(self):
        """Boolean specifying whether region is an HTAP binary region"""
        return True if self.region_id in HTAP_REGIONS else False

    def import_default(self, region_id):
        """Import region definition

        Parameters
        ----------
        region_id : str
            ID of region

        Raises
        ------
        KeyError
            if no region is registered for the input ID
        """
        self.update(REGION_DEFS[region_id])

        if self.lon_range_plot is None:
            self.lon_range_plot = self.lon_range
        if self.lat_range_plot is None:
            self.lat_range_plot = self.lat_range

    @property
    def center_coordinate(self):
        """Center coordinate of this region"""
        latc = self.lat_range[0] + (self.lat_range[1] - self.lat_range[0])/2
        lonc = self.lon_range[0] + (self.lon_range[1] - self.lon_range[0])/2
        return (latc, lonc)

    def distance_to_center(self, lat, lon):
        """Compute distance of input coordinate to center of this region

        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate

        Returns
        -------
        float
            distance in km
        """
        from pyaerocom.geodesy import calc_distance
        cc = self.center_coordinate
        return calc_distance(lat0=cc[0],
                             lon0=cc[1],
                             lat1=lat,
                             lon1=lon)

    def contains_coordinate(self, lat, lon):
        """Check if input lat/lon coordinate is contained in region

        Parameters
        ----------
        lat : float
            latitude of coordinate
        lon : float
            longitude of coordinate

        Returns
        -------
        bool
            True if coordinate is contained in this region, False if not
        """
        lat_ok = self.lat_range[0] <= lat <= self.lat_range[1]
        lon_ok = self.lon_range[0] <= lon <= self.lon_range[1]
        return lat_ok * lon_ok

    def mask_available(self):
        if not self.is_htap():
            return False
        return True

    def get_mask_data(self):
        if not self.mask_available():
            raise AttributeError(
                f'No binary mask data available for region {self.region_id}.'
                )
        if self._mask_data is None:
            self._mask_data = load_region_mask_xr(self.region_id)
        return self._mask_data

    def plot_mask(self, ax, color, alpha=0.2):

        mask = self.get_mask_data()
        #import numpy as np
        data = mask.data
        data[data==0] = np.nan
        mask.data = data

        mask.plot(ax=ax)
        return ax

    def plot_borders(self, ax, color, lw=2):
        raise NotImplementedError('Coming soon...')

    def plot(self, ax=None):
        """
        Plot this region

        Draws a rectangle of the outer bounds of the region and if a binary
        mask is available for this region, it will be plotted as well.

        Parameters
        ----------
        ax : GeoAxes, optional
            axes instance to be used for plotting. Defaults to None in which
            case a new instance is created.

        Returns
        -------
        GeoAxes
            axes instance used for plotting

        """
        from cartopy.mpl.geoaxes import GeoAxes
        from pyaerocom.plot.mapping import init_map
        if ax is None:
            ax = init_map()
        elif not isinstance(ax, GeoAxes):
            raise ValueError('Invalid input for ax: need cartopy GeoAxes..')

        if self.mask_available():
            self.plot_mask(ax, color='r')

        ax.set_xlabel('Longitude')
        ax.set_ylabel('Latitude')
        name = self.name
        if not name ==  self.region_id:
            name += f' (ID={self.region_id})'

        ax.set_title(name)

        return ax

    def __contains__(self, val):
        if not isinstance(val, tuple):
            raise TypeError('Invalid input, need tuple')
        if not len(val) == 2:
            raise ValueError('Invalid input: coordinate must contain 2 '
                             'elements (lat, lon)')
        return self.contains_coordinate(lat=val[0], lon=val[1])

    def __repr__(self):
       return ("Region %s %s" %(self.name, super(Region, self).__repr__()))

    def __str__(self):
        s = ("pyaeorocom Region\nName: %s\n"
             "Longitude range: %s\n"
             "Latitude range: %s\n"
             "Longitude range (plots): %s\n"
             "Latitude range (plots): %s"
             %(self.name, self.lon_range, self.lat_range,
               self.lon_range_plot, self.lat_range_plot))
        return s

def all():
    """Wrapper for :func:`get_all_default_region_ids`"""
    return list(REGION_DEFS)

def get_all_default_region_ids():
    """Get list containing IDs of all default regions

    Returns
    -------
    list
        IDs of all predefined default regions
    """
    return OLD_AEROCOM_REGIONS

def _get_regions_helper(reg_ids):
    """
    Get dictionary of :class:`Region` instances for input IDs

    Parameters
    ----------
    reg_ids : list
        list of region IDs

    Returns
    -------
    dict
        keys are input region IDs, values are loaded :class:`Region` instances
    """
    regs = {}
    for reg in reg_ids:
        regs[reg] = Region(reg)
    return regs

def get_old_aerocom_default_regions():
    """
    Load dictionary with default AeroCom regions

    Returns
    -------
    dict
        keys are region ID's, values are instances of :class:`Region`
    """
    return _get_regions_helper(OLD_AEROCOM_REGIONS)


def get_htap_regions():
    """
    Load dictionary with HTAP regions

    Returns
    -------
    dict
        keys are region ID's, values are instances of :class:`Region`
    """
    return _get_regions_helper(HTAP_REGIONS)

def get_all_default_regions():
    """Get dictionary containing all default regions from region.ini file

    Returns
    -------
    dict
        dictionary containing all default regions; keys are region ID's, values
        are instances of :class:`Region`.

    """
    return get_old_aerocom_default_regions()

#: ToDO: check how to handle methods properly with HTAP regions...
def get_regions_coord(lat, lon, regions=None):
    """Get all regions that contain input coordinate

    Note
    ----
    This does not yet include HTAP, since this causes troules in automated
    AeroCom processing

    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    regions : dict, optional
        dictionary containing instances of :class:`Region` as values, which
        are considered. If None, then all default regions are used.

    Returns
    -------
    list
        list of regions that contain this coordinate
    """

    matches = []
    if regions is None:
        regions = get_all_default_regions()
    for rname, reg in regions.items():
        if rname == 'WORLD': # always True
            continue
        if reg.contains_coordinate(lat, lon):
            matches.append(rname)
    if len(matches) == 0:
        matches.append('WORLD')
    return matches

def find_closest_region_coord(lat, lon, regions=None):
    """Find region that has it's center closest to input coordinate

    Parameters
    ----------
    lat : float
        latitude of coordinate
    lon : float
        longitude of coordinate
    regions : dict, optional
        dictionary containing instances of :class:`Region` as values, which
        are considered. If None, then all default regions are used.

    Returns
    -------
    str
        region ID of identified region
    """
    if regions is None:
        regions = get_all_default_regions()

    matches = get_regions_coord(lat, lon, regions)

    if len(matches) == 1:
        return matches[0]

    min_dist = 1e6
    best = None
    for match in matches:
        region = regions[match]
        dist = region.distance_to_center(lat, lon)
        if dist < min_dist:
            min_dist = dist
            best=match
    return best

if __name__=="__main__":
    import matplotlib.pyplot as plt
    plt.close('all')
    import pyaerocom as pya
    res = {}
    reg = Region('EUROPE')
    reg.plot()

# =============================================================================
#     for reg in pya.region_defs.HTAP_REGIONS:
#
#         mask = pya.helpers_landsea_masks.load_region_mask_xr(reg)
#         res[reg] = info = pya.helpers_landsea_masks.get_lat_lon_range_mask_region(mask)
#         lonr, latr = info['lon_range'], info['lat_range']
#
#
#         print(f'[{reg}]')
#         print(f'lon_range={lonr[0]:.3f},{lonr[1]:.3f}')
#         print(f'lat_range={latr[0]:.3f},{latr[1]:.3f}')
#         print()
# =============================================================================



import os
from getpass import getuser

from pyaerocom import const
from pyaerocom._lowlevel_helpers import (
    AsciiFileLoc,
    ConstrainedContainer,
    DirLoc,
    EitherOf,
    ListOfStrings,
    NestedContainer,
    StrType,
    read_json,
    write_json
)
from pyaerocom.aeroval.aux_io_helpers import ReadAuxHandler
from pyaerocom.aeroval.collections import ModelCollection, ObsCollection
from pyaerocom.aeroval.helpers import (
    _check_statistics_periods,
    _get_min_max_year_periods
)
from pyaerocom.colocation_auto import ColocationSetup
from pyaerocom.exceptions import AeroValConfigError


class OutputPaths(ConstrainedContainer):
    JSON_SUBDIRS = ['map', 'ts', 'ts/diurnal', 'scat', 'hm', 'hm/ts', 'contour']

    json_basedir = DirLoc(
        default=os.path.join(const.OUTPUTDIR, 'aeroval/data'),
        assert_exists=True,
        auto_create=True,
        logger=const.print_log,
        tooltip='Base directory for json output files')

    coldata_basedir = DirLoc(
        default=os.path.join(const.OUTPUTDIR, 'aeroval/coldata'),
        assert_exists=True,
        auto_create=True,
        logger=const.print_log,
        tooltip='Base directory for colocated data output files (NetCDF)')

    ADD_GLOB = ['coldata_basedir', 'json_basedir']
    proj_id = StrType()
    exp_id = StrType()
    def __init__(self, proj_id:str, exp_id:str,
                 json_basedir:str=None, coldata_basedir:str=None):
        self.proj_id = proj_id
        self.exp_id = exp_id
        if coldata_basedir:
            self.coldata_basedir = coldata_basedir
        if json_basedir:
            self.json_basedir = json_basedir

    def _check_init_dir(self, loc, assert_exists):
        if assert_exists and not os.path.exists(loc):
            os.makedirs(loc)
        return loc

    def get_coldata_dir(self, assert_exists=True):
        loc = os.path.join(self.coldata_basedir, self.proj_id, self.exp_id)
        return self._check_init_dir(loc, assert_exists)

    def get_json_output_dirs(self, assert_exists=True):
        out = {}
        base = os.path.join(self.json_basedir, self.proj_id, self.exp_id)
        for subdir in self.JSON_SUBDIRS:
            loc = self._check_init_dir(os.path.join(base, subdir),
                                       assert_exists)
            out[subdir] = loc
        return out

class ModelMapsSetup(ConstrainedContainer):
    maps_freq = EitherOf(['monthly', 'yearly'])
    def __init__(self, **kwargs):
        self.maps_res_deg = 5
        self.update(**kwargs)

class StatisticsSetup(ConstrainedContainer):
    """
    Setup options for statistical calculations

    Attributes
    ----------
    weighted_stats : bool
        if True, statistical parameters are calculated using area weights,
        this is only relevant for gridded / gridded evaluations.
    annual_stats_constrained : bool
        if True, then only sites are considered that satisfy a potentially
        specified annual resampling constraint (see
        :attr:`pyaerocom.colocation_auto.ColocationSetup.min_num_obs`). E.g.
        lets say you want to calculate statistical parameters (bias,
        correlation, etc.) for monthly model / obs data for a given site and
        year. Lets further say, that there are only 8 valid months of data, and
        4 months are missing, so statistics will be calculated for that year
        based on 8 vs. 8 values. Now if
        :attr:`pyaerocom.colocation_auto.ColocationSetup.min_num_obs` is
        specified in way that requires e.g. at least 9 valid months to
        represent the whole year, then this station will not be considered in
        case `annual_stats_constrained` is True, else it will. Defaults to
        False.
    stats_tseries_base_freq : str, optional
        The statistics Time Series display in AeroVal (under Overall Evaluation)
        is computed in intervals of a certain frequency, which is specified
        via :attr:`TimeSetup.main_freq` (defaults to monthly). That is,
        monthly colocated data is used as a basis to compute the statistics
        for each month (e.g. if you have 10 sites, then statistics will be
        computed based on 10 monthly values for each month of the timeseries,
        1 value for each site). `stats_tseries_base_freq` may be specified in
        case a higher resolution is supposed to be used as a basis to compute
        the timeseries in the resolution specified by
        :attr:`TimeSetup.main_freq` (e.g. if daily is specified here, then for
        the above example 310 values would be used - 31 for each site - to
        compute the statistics for a given month (in this case, a month with 31
        days, obviously).


    Parameters
    ----------
    kwargs
        any of the supported attributes, e.g.
        `StatisticsSetup(annual_stats_constrained=True)`

    """
    MIN_NUM = 1
    def __init__(self, **kwargs):
        self.weighted_stats = True
        self.annual_stats_constrained = False
        self.add_trends = False
        self.trends_min_yrs = 7
        self.stats_tseries_base_freq = None
        self.update(**kwargs)

class TimeSetup(ConstrainedContainer):
    DEFAULT_FREQS = ['monthly', 'yearly']
    SEASONS = ['all', 'DJF', 'MAM', 'JJA', 'SON']
    main_freq = StrType()
    freqs = ListOfStrings()
    periods = ListOfStrings()
    def __init__(self, **kwargs):
        self.main_freq = self.DEFAULT_FREQS[0]
        self.freqs = self.DEFAULT_FREQS
        self.add_seasons = True
        self.periods = []

    def get_seasons(self):
        """
        Get list of seasons to be analysed

        Returns :attr:`SEASONS` if :attr:`add_seasons` it True, else `[
        'all']` (only whole year).

        Returns
        -------
        list
            list of season strings for analysis

        """
        if self.add_seasons:
            return self.SEASONS
        return ['all']

    def _get_all_period_strings(self):
        """
        Get list of all period strings for evaluation

        Returns
        -------
        list
            list of period / season strings
        """
        output = []
        for per in self.periods:
            for season in self.get_seasons():
                perstr = f'{per}-{season}'
                output.append(perstr)
        return output

class WebDisplaySetup(ConstrainedContainer):

    map_zoom = EitherOf(['World', 'Europe'])
    regions_how = EitherOf(['default', 'aerocom', 'htap', 'country'])
    def __init__(self, **kwargs):
        self.regions_how = 'default'
        self.map_zoom = 'World'
        self.add_model_maps = False
        self.modelorder_from_config = True
        self.obsorder_from_config = True
        self.var_order_menu = []
        self.obs_order_menu = []
        self.model_order_menu = []
        self.update(**kwargs)

class EvalRunOptions(ConstrainedContainer):
    def __init__(self, **kwargs):
        # bool options run (do not affect results)
        self.clear_existing_json = True
        self.only_json = False
        self.only_colocation = False
        #: If True, process only maps (skip obs evaluation)
        self.only_model_maps = False
        self.update(**kwargs)

class ProjectInfo(ConstrainedContainer):
    def __init__(self, proj_id:str):
        self.proj_id = proj_id

class ExperimentInfo(ConstrainedContainer):
    def __init__(self, exp_id:str, **kwargs):
        self.exp_id = exp_id
        self.exp_name = ''
        self.exp_descr = ''
        self.public = False
        self.exp_pi = getuser()
        self.update(**kwargs)


class EvalSetup(NestedContainer, ConstrainedContainer):
    """Composite class representing a whole analysis setup

    This represents the level at which json I/O happens for configuration
    setup files.
    """
    IGNORE_JSON = ['_aux_funs']
    ADD_GLOB = ['io_aux_file']
    io_aux_file = AsciiFileLoc(
        default='',
        assert_exists=False,
        auto_create=False,
        logger=const.print_log,
        tooltip='.py file containing additional read methods for modeldata')
    _aux_funs = {}
    def __init__(self, proj_id:str=None, exp_id:str=None, **kwargs):
        if proj_id is None:
            proj_id = kwargs['proj_info']['proj_id']
        if exp_id is None:
            exp_id = kwargs['exp_info']['exp_id']

        self.proj_info = ProjectInfo(proj_id=proj_id)
        self.exp_info = ExperimentInfo(exp_id=exp_id)

        self.time_cfg = TimeSetup()

        self.modelmaps_opts = ModelMapsSetup()
        self.colocation_opts = ColocationSetup(
                                    save_coldata=True,
                                    keep_data=False,
                                    resample_how='mean'
                                    )
        self.statistics_opts = StatisticsSetup(
                                    weighted_stats=True,
                                    annual_stats_constrained=False
                                    )
        self.webdisp_opts = WebDisplaySetup()

        self.processing_opts = EvalRunOptions()

        self.obs_cfg = ObsCollection()
        self.model_cfg = ModelCollection()

        self.var_web_info = {}
        self.path_manager = OutputPaths(self.proj_id, self.exp_id)
        self.update(**kwargs)

    @property
    def proj_id(self) -> str:
        """
        str: proj ID (wrapper to :attr:`proj_info.proj_id`)
        """
        return self.proj_info.proj_id

    @property
    def exp_id(self) -> str:
        """
        str: experiment ID (wrapper to :attr:`exp_info.exp_id`)
        """
        return self.exp_info.exp_id

    @property
    def json_filename(self) -> str:
        """
        str: Savename of config file: cfg_<proj_id>_<exp_id>.json
        """
        return f'cfg_{self.proj_id}_{self.exp_id}.json'

    @property
    def gridded_aux_funs(self):
        if not bool(self._aux_funs) and os.path.exists(self.io_aux_file):
            self._import_aux_funs()
        return self._aux_funs

    def get_obs_entry(self, obs_name):
        return self.obs_cfg.get_entry(obs_name).to_dict()

    def get_model_entry(self, model_name):
        """Get model entry configuration

        Since the configuration files for experiments are in json format, they
        do not allow the storage of executable custom methods for model data
        reading. Instead, these can be specified in a python module that may
        be specified via :attr:`add_methods_file` and that contains a
        dictionary `FUNS` that maps the method names with the callable methods.

        As a result, this means that, by default, custom read methods for
        individual models in :attr:`model_config` do not contain the
        callable methods but only the names. This method will take care of
        handling this and will return a dictionary where potential custom
        method strings have been converted to the corresponding callable
        methods.

        Parameters
        ----------
        model_name : str
            name of model

        Returns
        -------
        dict
            Dictionary that specifies the model setup ready for the analysis
        """
        cfg = self.model_cfg.get_entry(model_name)
        cfg = cfg.prep_dict_analysis(self.gridded_aux_funs)
        return cfg

    def to_json(self, outdir:str, ignore_nan:bool=True, indent:int=3) -> None:
        """
        Save configuration as JSON file

        Parameters
        ----------
        outdir : str
            directory where the config json file is supposed to be stored
        ignore_nan : bool
            set NaNs to Null when writing
        indent : int
            json indentation

        """
        filepath = os.path.join(outdir, self.json_filename)
        data = self.json_repr()
        write_json(data, filepath,
                   ignore_nan=ignore_nan,
                   indent=indent)
        return filepath

    @staticmethod
    def from_json(filepath:str) -> 'EvalSetup':
        """Load configuration from json config file"""
        settings = read_json(filepath)
        return EvalSetup(**settings)

    def _import_aux_funs(self):

         h = ReadAuxHandler(self.io_aux_file)
         self._aux_funs.update(**h.import_all())

    def _check_time_config(self):
        periods = self.time_cfg.periods
        colstart = self.colocation_opts['start']
        colstop = self.colocation_opts['stop']

        if len(periods) == 0:
            if colstart is None:
                raise AeroValConfigError(
                    'Either periods or start must be set...'
                    )
            per = self.colocation_opts._period_from_start_stop()
            periods = [per]
            const.print_log.info(
                f'periods is not set, inferred {per} from start '
                f'/ stop colocation settings.')

        self.time_cfg['periods'] = _check_statistics_periods(periods)
        start, stop = _get_min_max_year_periods(periods)
        if colstart is None:
            self.colocation_opts['start'] = start
        if colstop is None:
            self.colocation_opts['stop'] = stop + 1 # add 1 year since we want to include stop year






import os
from configparser import ConfigParser

from pyaerocom.exceptions import VariableDefinitionError


def parse_variables_ini(fpath=None):
    """Returns instance of ConfigParser to access information"""
    from pyaerocom import __dir__
    if fpath is None:
        fpath = os.path.join(__dir__, "data", "variables.ini")

    if not os.path.exists(fpath):
        raise FileNotFoundError("FATAL: variables.ini file could not be found "
                                "at {}".format(fpath))
    parser = ConfigParser()
    parser.read(fpath)
    return parser


def parse_aliases_ini():
    """Returns instance of ConfigParser to access information"""
    from pyaerocom import __dir__
    fpath = os.path.join(__dir__, "data", "aliases.ini")
    if not os.path.exists(fpath):
        raise FileNotFoundError("FATAL: aliases.ini file could not be found "
                                "at {}".format(fpath))
    parser = ConfigParser()
    parser.read(fpath)
    return parser


def get_emep_variables(parser=None):
    """Read variable definitions from emep_variables.ini file

    Returns
    -------
    dict
        keys are AEROCOM standard names of variable, values are EMEP variables
    """
    if parser is None:
        parser = parse_emep_variables_ini()
    variables = {}
    items = parser['emep_variables']
    for var_name in items:
        _variables = [x.strip() for x in items[var_name].strip().split(',')]
        for variable in _variables:
            variables[var_name] = variable
    return variables


def parse_emep_variables_ini(fpath=None):
    """Returns instance of ConfigParser to access information"""
    from pyaerocom import __dir__
    if fpath is None:
        fpath = os.path.join(__dir__, "data", "emep_variables.ini")
    if not os.path.exists(fpath):
        raise FileNotFoundError("FATAL: emep_variables.ini file could not be found "
                        "at {}".format(fpath))
    parser = ConfigParser()
    # added 12.7.21 by jgliss for EMEP trends processing. See here:
    # https://stackoverflow.com/questions/1611799/preserve-case-in-configparser
    parser.optionxform=str
    parser.read(fpath)
    return parser


def _read_alias_ini(parser=None):
    """Read all alias definitions from aliases.ini file and return as dict

    Returns
    -------
    dict
        keys are AEROCOM standard names of variable, values are corresponding
        aliases
    """
    if parser is None:
        parser = parse_aliases_ini()
    aliases = {}
    items = parser['aliases']
    for var_name in items:
        _aliases = [x.strip() for x in items[var_name].strip().split(',')]
        for alias in _aliases:
            aliases[alias] = var_name
    for var_fam, alias_fam in parser['alias_families'].items():
        if ',' in alias_fam:
            raise Exception('Found invalid definition of alias family {}: {}. '
                            'Only one family can be mapped to a variable name'
                            .format(var_fam, alias_fam))
    return aliases


def get_aliases(var_name, parser=None):
    """Get aliases for a certain variable"""
    if parser is None:
        from pyaerocom import __dir__
        file = os.path.join(__dir__, "data", "aliases.ini")
        parser = ConfigParser()
        parser.read(file)

    info = parser['aliases']
    aliases = []
    if var_name in info:
        aliases.extend([a.strip() for a in info[var_name].split(',')])
    for var_fam, alias_fam in parser['alias_families'].items():
        if var_name.startswith(var_fam):
            alias = var_name.replace(var_fam, alias_fam)
            aliases.append(alias)
    return aliases


def _check_alias_family(var_name, parser):
    for var_fam, alias_fam in parser['alias_families'].items():
        if var_name.startswith(alias_fam):
            var_name_aerocom = var_name.replace(alias_fam, var_fam)
            return var_name_aerocom
    raise VariableDefinitionError('Input variable could not be identified as '
                                  'belonging to either of the available alias '
                                  'variable families')


def get_variable(var_name):
    """
    Get a certain variable

    Parameters
    ----------
    var_name : str
        variable name

    Returns
    -------
    Variable
    """
    from pyaerocom import const
    return const.VARS[var_name]


from pyaerocom.aeroval.varinfo_web import VarinfoWeb
from pyaerocom.helpers import start_stop_str
from pyaerocom.colocation_auto import Colocator
from pyaerocom.tstype import TsType
from pyaerocom.exceptions import TemporalResolutionError


def check_var_ranges_avail(model_data, var_name):
    """
    Check if lower and upper variable ranges are available for input variable

    Parameters
    ----------
    model_data : GriddedData
        modeldata containing variable data
    var_name : str
        variable name to be checked (must be the same as model data
        AeroCom variable name).

    Raises
    ------
    ValueError
        if ranges for input variable are not defined and if input model data
        corresponds to a different variable than the input variable name.

    Returns
    -------
    None

    """
    try:
        VarinfoWeb(var_name)
    except AttributeError:
        if model_data.var_name_aerocom == var_name:
            model_data.register_var_glob(delete_existing=True)
        else:
            raise ValueError(
                f'Mismatch between variable name of input model_data '
                f'({model_data.var_name_aerocom}) and var_name {var_name}')


def _check_statistics_periods(periods: list) -> list:
    """
    Check input list of period strings is valid

    Parameters
    ----------
    periods : list
        list containing period strings to be checked.

    Raises
    ------
    ValueError
        if input is not a list or any of the provided periods in that list is
        not a string or invalid.

    Returns
    -------
    list
        list of periods

    """
    checked = []
    if not isinstance(periods, list):
        raise ValueError('statistics_periods needs to be a list')
    for per in periods:
        if not isinstance(per, str):
            raise ValueError('All periods need to be strings')
        spl = [x.strip() for x in per.split('-')]
        if len(spl) > 2:
            raise ValueError(
                f'Invalid value for period ({per}), can be either single '
                f'years or period of years (e.g. 2000-2010).'
                )
        _per = '-'.join([str(int(val)) for val in spl])
        checked.append(_per)
    return checked

def _period_str_to_timeslice(period: str) -> slice:
    """
    Convert input period to a time slice

    Parameters
    ----------
    period : str
        period, e.g. "2000-2010"

    Raises
    ------
    ValueError
        if input period is invalid

    Returns
    -------
    slice
        slice containing start and end strings.
    """
    spl = period.split('-')
    if len(spl) == 1:
        return slice(spl[0],spl[0])
    elif len(spl) == 2:
        return slice(*spl)
    raise ValueError(period)

def _get_min_max_year_periods(statistics_periods):
    """Get lowest and highest available year from all periods

    Parameters
    ----------
    statistics_periods : list
        list of periods for experiment

    Returns
    -------
    int
        start year
    int
        stop year (may be the same as start year, e.g. if periods suggest
        single year analysis).
    """
    startyr, stopyr = 1e6, -1e6
    for per in statistics_periods:
        sl = _period_str_to_timeslice(per)
        perstart, perstop = int(sl.start), int(sl.stop)
        if perstart < startyr:
            startyr = perstart
        if perstop > stopyr:
            stopyr = perstop
    return startyr, stopyr


import numpy as np
from cf_units import Unit

from pyaerocom.tstype import TsType
from pyaerocom.exceptions import UnitConversionError
from pyaerocom.helpers import isnumeric, resample_time_dataarray
from pyaerocom.time_config import SI_TO_TS_TYPE
from pyaerocom.units_helpers import get_unit_conversion_fac

# ToDo: check if still needed
DEP_IMPLICIT_UNITS = [Unit('mg N m-2'),
                      Unit('mg S m-2'),
                      Unit('mg m-2')]
PR_IMPLICIT_UNITS = [Unit('mm')]
DEP_TEST_UNIT = 'kg m-2 s-1'

# ToDo: check if still needed
def translate_rate_units_implicit(unit_implicit, ts_type):
    unit = Unit(unit_implicit)

    freq = TsType(ts_type)
    freq_si = freq.to_si()

    # check if unit is explicitly defined as implicit and if yes add frequency
    # string
    found = False
    for imp_unit in DEP_IMPLICIT_UNITS:
        if unit == imp_unit:
            unit = f'{imp_unit} {freq_si}-1'
            found = True
            break

    # Check if frequency in unit corresponds to sampling frequency (e.g.
    # ug m-2 h-1 for hourly data).
    freq_si_str = f'{freq_si}-1'
    freq_si_str_alt = f'/{freq_si}'
    if str(unit).endswith(freq_si_str_alt):
        # make sure frequency is denoted as e.g. m s-1 instead of m/s
        _new =str(unit).replace(freq_si_str_alt, freq_si_str)
        unit = Unit(_new)

    # for now, raise NotImplementedError if wdep unit is, e.g. ug m-2 s-1 but
    # ts_type is hourly (later, use units_helpers.implicit_to_explicit_rates)
    if not freq_si_str in str(unit):
        raise NotImplementedError(f'Cannot yet handle wdep in {unit} but '
                                  f'{freq} sampling frequency')
    return unit

def _check_unit_conversion_fac(unit, test_unit, non_si_info=None): # pragma: no cover
    raise NotImplementedError('under revision')
    if non_si_info is None:
        non_si_info = []
    try:
        get_unit_conversion_fac(unit, DEP_TEST_UNIT)
        return True
    except UnitConversionError:
        for substr in non_si_info:
            if substr in unit:
                check = unit.replace(substr, '')
                return _check_unit_conversion_fac(check, test_unit)
    return False

# ToDo: check if still needed
def check_pr_units(gridded): # pragma: no cover
    raise NotImplementedError('under revision')
    # ToDo: harmonise input and output with translate_rate_units_implicit
    unit = Unit(gridded.units)
    freq = TsType(gridded.ts_type)
    freq_si = freq.to_si()

    # check if precip unit is implicit
    if any([unit == x for x in PR_IMPLICIT_UNITS]):
        unit = f'{unit} {freq_si}-1'
        gridded.units = unit

    # Check if frequency in unit corresponds to sampling frequency (e.g.
    # ug m-2 h-1 for hourly data).
    freq_si_str = f' {freq_si}-1'
    freq_si_str_alt = f'/{freq_si}'
    if freq_si_str_alt in str(unit):
        # make sure frequencey is denoted as e.g. m s-1 instead of m/s
        unit = str(unit).replace(freq_si_str_alt,
                                 freq_si_str)

        gridded.units = unit

    # for now, raise NotImplementedError if wdep unit is, e.g. ug m-2 s-1 but
    # ts_type is hourly (later, use units_helpers.implicit_to_explicit_rates)
    if not freq_si_str in str(unit):
        raise NotImplementedError(f'Cannot yet handle wdep in {unit} but '
                                  f'{freq} sampling frequency')
    return gridded

# ToDo: check if still needed
def _check_prlim_units(prlim, prlim_units): # pragma: no cover
    raise NotImplementedError('under revision')
    # ToDo: cumbersome for now, make it work first, then make it simpler...
    if not prlim_units.endswith('-1'):
        raise ValueError('Please specify prlim_unit as string ending with '
                         '-1 (e.g. mm h-1) or similar')

    spl = prlim_units.split()
    if not len(spl) == 2:
        raise ValueError('Invalid input for prlim_units (only one whitespace '
                         'is allowed)')
    # make sure to be in the correct length unit
    mulfac = get_unit_conversion_fac(spl[0], 'm')
    prlim *= mulfac

    prlim_units = f'm {spl[1]}'
    prlim_freq = spl[1][:-2]  # it endswith -1
    # convert the freque
    if not prlim_freq in SI_TO_TS_TYPE:
        raise ValueError(
            f'frequency in prlim_units must be either of the '
            f'following values: {list(SI_TO_TS_TYPE.keys())}.')

    prlim_tstype = TsType(SI_TO_TS_TYPE[prlim_freq])
    return (prlim, prlim_units, prlim_tstype)

# ToDo: check if still needed
def _apply_prlim_wdep(wdeparr, prarr, prlim, prlim_unit, prlim_set_under): # pragma: no cover
    raise NotImplementedError('under revision')
    if prlim_unit is None:
        raise ValueError(f'Please provide prlim_unit for prlim={prlim}')
    elif prlim_set_under is None:
        raise ValueError(f'Please provide prlim_set_under for prlim={prlim}')
    elif not isnumeric(prlim_set_under):
        raise ValueError(f'Please provide a numerical value or np.nan for '
                         f'prlim_set_under, got {prlim_set_under}')

    prmask = prarr.data < prlim
    wdeparr.data[prmask] = prlim_set_under

    return wdeparr, prmask

# ToDo: check if still needed
def _aggregate_wdep_pr(wdeparr, prarr, wdep_unit, pr_unit, from_tstype,
                       to_tstype): # pragma: no cover
    raise NotImplementedError('under revision')
    to_tstype_pd = to_tstype.to_pandas_freq()
    to_tstype_si = to_tstype.to_si()

    from_tstype_si = from_tstype.to_si()

    wdeparr = resample_time_dataarray(wdeparr,
                                      to_tstype_pd,
                                      how='sum')
    wdep_unit = wdep_unit.replace(f'{from_tstype_si}-1',
                                  f'{to_tstype_si}-1')
    prarr = resample_time_dataarray(prarr,
                                    to_tstype_pd,
                                    how='sum')
    pr_unit = pr_unit.replace(f'{from_tstype_si}-1',
                              f'{to_tstype_si}-1')

    return (wdeparr, prarr, wdep_unit, pr_unit)

# ToDo: check if still needed
def compute_concprcp_from_pr_and_wetdep(wdep, pr, ts_type=None,
                                        prlim=None, prlim_units=None,
                                        prlim_set_under=None): # pragma: no cover
    raise NotImplementedError('under revision')
    if ts_type is None:
        ts_type = wdep.ts_type

    wdep_tstype = TsType(wdep.ts_type)

    # get units from deposition input and precipitation; sometimes, they are
    # defined implicit, e.g. mm for precipitation, which is then already
    # accumulated over the provided time resolution in the data, that is, if
    # the data is hourly and precip is in units of mm, then it means the the
    # unit is mm/h. In addition, wet deposition units may be in mass of main
    # atom (e.g. N, or S) which are not SI and thus, not handled properly by
    # CF units.
    wdep_unit = str(wdep.units)

    wdep = translate_rate_units_implicit(wdep_unit,
                                         wdep_tstype)
    pr = check_pr_units(pr)

    # repeat the unit check steps done for wet deposition
    pr_unit = str(pr.units)
    pr_tstype = TsType(pr.ts_type)

    if not wdep_tstype == pr_tstype:
        # ToDo: this can probably fixed via time resampling with how='sum'
        # for the higher resolution dataset, but for this first draft, this
        # is not allowed.
        raise ValueError('Input precipitation and wet deposition fields '
                         'need to be in the same frequency...')

    # assign input frequency (just for making the code better readable)
    from_tstype = wdep_tstype
    from_tstype_si = from_tstype.to_si()

    # convert data objects to xarray to do modifications and computation of
    # output variable
    prarr = pr.to_xarray()
    wdeparr = wdep.to_xarray()

    # Make sure precip unit is correct for concprcp=wdep/pr
    pr_unit_calc = f'm {from_tstype_si}-1'

    # unit conversion factor for precip
    mulfac_pr = get_unit_conversion_fac(pr_unit, pr_unit_calc)

    if mulfac_pr != 1:
        prarr *= mulfac_pr
        pr_unit = pr_unit_calc

    # Make sure wdep unit is correct for concprcp=wdep/pr
    wdep_unit_check_calc = wdep_unit.replace('N', '').replace('S', '')
    wdep_unit_calc = f'ug m-2 {from_tstype_si}-1'
    mulfac_wdep = get_unit_conversion_fac(wdep_unit_check_calc, wdep_unit_calc)

    if mulfac_wdep != 1:
        wdeparr *= mulfac_wdep
        if 'N' in wdep_unit:
            wdep_unit_calc = wdep_unit_calc.replace('ug', 'ug N')
        elif 'S' in wdep_unit:
            wdep_unit_calc = wdep_unit_calc.replace('ug', 'ug S')
        wdep_unit = wdep_unit_calc

    # final output frequency (precip limit may be applied in higher resolution)
    to_tstype = TsType(ts_type)
    to_tstype_si = to_tstype.to_si()

    # apply prlim filter if applicable
    apply_prlim, prlim_applied = False, False
    if prlim is not None:
        apply_prlim = True

        prlim, prlim_units, prlim_tstype = _check_prlim_units(prlim,
                                                              prlim_units)

        if prlim_tstype == from_tstype:
            wdeparr, _ = _apply_prlim_wdep(wdeparr, prarr, prlim, prlim_units,
                                           prlim_set_under)
            prlim_applied = True

    if apply_prlim and not prlim_applied and prlim_tstype > to_tstype:
        # intermediate frequency where precip filter should be applied
        (wdeparr,
         prarr,
         wdep_unit,
         pr_unit) = _aggregate_wdep_pr(wdeparr, prarr, wdep_unit, pr_unit,
                                       from_tstype, prlim_tstype)

        wdeparr, _ = _apply_prlim_wdep(wdeparr, prarr, prlim, prlim_units,
                                       prlim_set_under)
        prlim_applied = True
        from_tstype = prlim_tstype

    if not from_tstype == to_tstype:
        (wdeparr,
         prarr,
         wdep_unit,
         pr_unit) = _aggregate_wdep_pr(wdeparr, prarr, wdep_unit, pr_unit,
                                       from_tstype, to_tstype)

    if apply_prlim and not prlim_applied:
        if not to_tstype == prlim_tstype:
            raise ValueError('... ... .. ')
        wdeparr, _ = _apply_prlim_wdep(wdeparr, prarr,
                                       prlim, prlim_units,
                                       prlim_set_under)

    # set PR=0 to NaN (as we divide py PR)
    prarr.data[prarr.data == 0] = np.nan

    concprcparr = wdeparr / prarr

    cube = concprcparr.to_iris()
    # infer output unit of concentration variable (should be ug m-3 or ug N m-3 or ug S m-3)
    conc_unit_out = wdep_unit.replace('m-2', 'm-3').replace(
        f'{to_tstype_si}-1', '').strip()
    cube.units = conc_unit_out
    cube.attributes['ts_type'] = str(to_tstype)
    return cube



from pyaerocom.mathutils import exponent
import numpy as np
import cartopy.crs as ccrs

def projection_from_str(projection_str="PlateCarree"):
    """Return instance of cartopy projection class based on string ID"""
    try:
        return ccrs.__dict__[projection_str]()
    except KeyError:
        raise ValueError(f'no such projection {projection_str}')

def custom_mpl(mpl_rcparams=None, default_large=True, **kwargs):
    """Custom matplotlib settings"""
    if mpl_rcparams is None:
        from matplotlib import rcParams as mpl_rcparams
    small = 10
    medium = 12
    big = 14
    huge = 18
    if default_large:
        default = {'font.size'          :   huge,
                   'axes.titlesize'     :   huge,
                   'axes.labelsize'     :   huge,
                   'xtick.labelsize'    :   big,
                   'ytick.labelsize'    :   big,
                   'legend.fontsize'    :   big,
                   'figure.titlesize'   :   huge}
    else:
        default = {'font.size'          :   small,
                   'axes.titlesize'     :   big,
                   'axes.labelsize'     :   medium,
                   'xtick.labelsize'    :   medium,
                   'ytick.labelsize'    :   medium,
                   'legend.fontsize'    :   small,
                   'figure.titlesize'   :   big}

    for k, v in default.items():
        try:
            mpl_rcparams[k] = kwargs[k]
        except Exception:
            mpl_rcparams[k] = v
    return mpl_rcparams

def calc_figsize(lon_range, lat_range, figh=8):
    """Calculate figure size based on data

    The required figure width is computed based on the input height and the
    aspect ratio of the longitude and latitude arrays

    Parameters
    ----------
    lon_range : tuple
        2-element tuple specifying longitude range (may also be list or array)
    lat_range : tuple
        2-element tuple specifying latitude range (may also be list or array)
    figh : int
        figure height in inches
    add_cbar : bool
        if True, the width is adapted accordingly

    Returns
    -------
    tuple
        2-element tuple containing figure width and height
    """
    wfac = (lon_range[1] - lon_range[0]) / (lat_range[1] - lat_range[0])
    figw = int(wfac*figh)

    return (figw, figh)

def calc_pseudolog_cmaplevels(vmin, vmax, add_zero=False):
    """Initiate pseudo-log discrete colormap levels

    Parameters
    ----------
    vmin : float
        lower end of colormap (e.g. minimum value of data)
    vmax : float
        upper value of colormap (e.g. maximum value of data)
    add_zero : bool
        if True, the lower bound is set to 0 (irrelevant if vmin is 0).

    Returns
    -------
    list
        list containing boundary array for discrete colormap (e.g. using
        BoundaryNorm)

    Example
    -------
    >>> vmin, vmax = 0.02, 0.75
    >>> vals = calc_pseudolog_cmaplevels(vmin, vmax, num_per_mag=10, add_zero=True)
    >>> for val in vals: print("%.4f" %val)
    0.0000
    0.0100
    0.0126
    0.0158
    0.0200
    0.0251
    0.0316
    0.0398
    0.0501
    0.0631
    0.0794
    0.1000

    """

    if vmin < 0:
        vmin = 0
    if vmin == 0:
        vmin = 1*10.0**(exponent(vmax) - 2)
        if not add_zero:
            add_zero = True
    elif vmax < vmin:
        raise ValueError("Error: vmax must exceed vmin")
    bounds = [0] if add_zero else []
    low =  float(exponent(vmin))
    high = float(exponent(vmax))
    bounds.extend(np.arange(np.floor(vmin*10**(-low)), 10, 1)*10.0**(low))
    if low == high:
        return bounds

    for mag in range(int(low+1), int(high)):
        bounds.extend(np.linspace(1,9,9)*10**(mag))
    bounds.extend(np.arange(1, np.ceil(vmax*10**(-high)), 1)*10.0**(high))
    bounds.append(vmax)
    return bounds

def get_cmap_levels_auto(vmin, vmax, num_per_mag=10):
    """Initiate pseudo-log discrete colormap levels

    Note
    ----
        This is a beta version and aims to

    Parameters
    ----------
    vmin : float
        lower end of colormap (e.g. minimum value of data)
    vmax : float
        upper value of colormap (e.g. maximum value of data)
    """
    high = float(exponent(vmax))
    low = -3. if vmin == 0 else float(exponent(vmin))
    lvls =[]
    if 1%vmax*10**(-high) == 0:
        low+=1
        high-=1
    for mag in range(int(low), int(high)):
        lvls.extend(np.linspace(1, 10, num_per_mag-1, endpoint=0)*10**(mag))
    lvls.extend(np.linspace(1, vmax*10**(-high), num_per_mag, endpoint=1)*10**(high))

    return lvls

def get_cmap_ticks_auto(lvls, num_per_mag=3):
    """Compute cmap ticks based on cmap levels

    The cmap levels may be computed automatically using
    :func:`get_cmap_levels_auto`.

    Parameters
    ----------
    lvls : list
        list containing colormap levels
    num_per_mag : int
        desired number of ticks per magnitude
    """
    low = exponent(lvls[1]) # second entry (first is 0)
    vmax = lvls[-1]
    high = exponent(vmax)

    ticks = [0]
    if 1%vmax*10**(-high) == 0:
        for mag in range(low, high-1):
            ticks.extend(np.linspace(1, 10, num_per_mag, endpoint=0)*10**(mag))
        ticks.extend(np.linspace(1, 10, num_per_mag+1, endpoint=1)*10**(high-1))
    else:
        for mag in range(low, high):
            ticks.extend(np.linspace(1, 10, num_per_mag, endpoint=0)*10**(mag))
        ticks.extend(np.linspace(1, vmax*10**(-high),
                     num_per_mag+1, endpoint=1)*10**(high))

    return ticks

if __name__=="__main__":

    vmin, vmax = 0.02, 0.75

    vals = calc_pseudolog_cmaplevels(vmin, vmax, num_per_mag=10, add_zero=True)
    for val in vals: print("%.4f" %val)

    import doctest
    doctest.testmod()



"""
Settings and helper methods / classes for I/O of obervation data

Note
----
Some settings like paths etc can be found in :mod:`pyaerocom.config.py`
"""
from pyaerocom._lowlevel_helpers import str_underline, dict_to_str
#: Wavelength tolerance for observations if data for required wavelength
#: is not available
OBS_WAVELENGTH_TOL_NM = 10.0

#: This boolean can be used to enable / disable the former (i.e. use
#: available wavelengths of variable in a certain range around variable
#: wavelength).
OBS_ALLOW_ALT_WAVELENGTHS = True


class ObsVarCombi(object):
    def __init__(self, obs_id, var_name):
        self.obs_id = obs_id
        self.var_name = var_name

    def __repr__(self):
        return '{};{}'.format(self.obs_id, self.var_name)

    def __str__(self):
        return repr(self)

class AuxInfoUngridded(object):
    MAX_VARS_PER_METHOD = 2
    def __init__(self, data_id, vars_supported, aux_requires,
                 aux_merge_how, aux_funs=None, aux_units=None):

        self.data_id = data_id
        self.vars_supported = vars_supported

        self.aux_requires = aux_requires
        self.aux_merge_how = aux_merge_how
        self.aux_funs = aux_funs
        self.aux_units = aux_units
        self.check_status()

    def to_dict(self):
        """Dictionary representation of this object

        Ignores any potential private attributes.
        """
        dd = {}
        for key, val in self.__dict__.items():
            if any([key.startswith(x) for x in ('_', '__')]):
                continue
            dd[key] = val
        return dd

    def check_status(self):
        """
        Check if specifications are correct and consistent

        Raises
        ------
        ValueError
            If one of the class attributes is invalid
        NotImplementedError
            If computation method contains more than 2 variables / datasets

        """
        if isinstance(self.vars_supported, str):
            self.vars_supported = [self.vars_supported]

        if isinstance(self.aux_merge_how, str):
            nv = len(self.vars_supported)
            self.aux_merge_how = dict(zip(self.vars_supported,
                                          [self.aux_merge_how]*nv))
        if self.aux_funs is None:
            self.aux_funs = {}
        if self.aux_units is None:
            self.aux_units ={}

        for var in self.vars_supported:
            if not var in self.aux_requires:
                raise ValueError('Variable {} is not defined in attr '
                                 'aux_requires...'.format(var))

            elif not var in self.aux_merge_how:
                raise ValueError('Missing information about how {} should '
                                 'be merged (aux_merge_how)'.format(var))
            merge_how = self.aux_merge_how[var]
            if merge_how == 'eval':
                if not var in self.aux_funs:
                    raise ValueError('Specification of computation function is '
                                     'missing for var {}'.format(var))
                fun = self.aux_funs[var]

                if not isinstance(fun, str):
                    raise ValueError('eval functions need to be strings')

            aux_info = self.aux_requires[var]

            fc = 0
            for aux_id, var_info in aux_info.items():
                if isinstance(var_info, str):
                    # make sure variables are represented as list, even if
                    # it is only one
                    aux_info[aux_id] = var_info = [var_info]
                for _var in var_info:
                    obsvar = ObsVarCombi(aux_id, _var)
                    obsvarstr = str(obsvar)
                    if merge_how == 'eval' and not obsvarstr in fun:
                        raise ValueError('Mismatch between aux_requires and '
                                         'aux_funs for variable {}. No such '
                                         'obs;var string {} in computation '
                                         'method {}'
                                         .format(var, obsvarstr, fun))

                    fc += 1
                    if fc > self.MAX_VARS_PER_METHOD:
                        raise NotImplementedError('So far only 2 variables '
                                                  'can be combined...')

    def __repr__(self):
        return ('{}; data_id: {}; vars_supported: {}'
                .format(str(type(self).__name__), self.data_id, self.vars_supported))

    def __str__(self):
        name = str(type(self).__name__)
        s = str_underline(name)
        s += dict_to_str(self.to_dict())
        return s

if __name__ == '__main__':

    print(ObsVarCombi('Bla', 'od550aer'))

    info = AuxInfoUngridded('bla',
                            'od550aer',
                            {'od550aer': {'a':'od550lt1aer',
                                          'b':'od550gt1aer'}},
                            aux_merge_how='combine',
                            aux_funs=None, aux_units=None)

    print(info)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 29 17:12:54 2020

@author: jonasg
"""

from pathlib import Path
import os
import requests
import tarfile

from traceback import format_exc
from pyaerocom import const
from pyaerocom.io import (ReadAeronetSunV3, ReadAeronetSdaV3, ReadAeronetInvV3,
                          ReadEbas, ReadAirNow, ReadGhost, ReadEEAAQEREP_V2,
                          ReadEarlinet)

class AccessTestData:

    #: That's were the testdata can be downloaded from
    URL_TESTDATA = 'https://pyaerocom.met.no/pyaerocom-suppl/testdata-minimal.tar.gz'

    #: Directory where testdata will be downloaded into
    BASEDIR_DEFAULT = const.OUTPUTDIR

    #: Name of testdata directory
    TESTDATADIRNAME = 'testdata-minimal'

    #: Paths to be added to pya.const. All relative to :attr:`basedir`
    ADD_PATHS = {

        'MODELS'                        : 'modeldata',
        'OBSERVATIONS'                  : 'obsdata',
        'CONFIG'                        : 'config',
        'AeronetSunV3L2Subset.daily'    : 'obsdata/AeronetSunV3Lev2.daily/renamed',
        'AeronetSDAV3L2Subset.daily'    : 'obsdata/AeronetSDAV3Lev2.daily/renamed',
        'AeronetInvV3L2Subset.daily'    : 'obsdata/AeronetInvV3Lev2.daily/renamed',
        'EBASSubset'                    : 'obsdata/EBASMultiColumn',
        'AirNowSubset'                  : 'obsdata/AirNowSubset',
        'G.EEA.daily.Subset'            : 'obsdata/GHOST/data/EEA_AQ_eReporting/daily',
        'G.EEA.hourly.Subset'           : 'obsdata/GHOST/data/EEA_AQ_eReporting/hourly',
        'G.EBAS.daily.Subset'           : 'obsdata/GHOST/data/EBAS/daily',
        'G.EBAS.hourly.Subset'          : 'obsdata/GHOST/data/EBAS/hourly',
        'EEA_AQeRep.v2.Subset'          : 'obsdata/EEA_AQeRep.v2/renamed',
        'Earlinet-test'                 : 'obsdata/Earlinet'


    }

    _UNGRIDDED_READERS = {
        'AeronetSunV3L2Subset.daily'  : ReadAeronetSunV3,
        'AeronetSDAV3L2Subset.daily'  : ReadAeronetSdaV3,
        'AeronetInvV3L2Subset.daily'  : ReadAeronetInvV3,
        'EBASSubset'                  : ReadEbas,
        'AirNowSubset'                : ReadAirNow,
        'G.EEA.daily.Subset'          : ReadGhost,
        'G.EEA.hourly.Subset'         : ReadGhost,
        'G.EBAS.daily.Subset'         : ReadGhost,
        'G.EBAS.hourly.Subset'        : ReadGhost,
        'EEA_AQeRep.v2.Subset'        : ReadEEAAQEREP_V2,
        'Earlinet-test'               : ReadEarlinet

    }

    def __init__(self, basedir=None):
        self._basedir = None
        if basedir is not None:
            self.basedir = basedir

    @property
    def basedir(self):
        """Directory into which testdata will be downloaded"""
        if self._basedir is not None:
            return Path(self._basedir)
        return Path(self.BASEDIR_DEFAULT)

    @basedir.setter
    def basedir(self, val):
        if not os.path.isdir(val):
            raise ValueError(f'Input basedir {val} is not a directory...')
        self._basedir = val

    @property
    def testdatadir(self):
        """Directory containing testdata"""
        return self.basedir.joinpath(self.TESTDATADIRNAME)

    def download(self, basedir=None):
        """
        Download testdata

        Parameters
        ----------
        basedir : str, optional
            directory in which testdata is supposed to be downloaded
        Returns
        -------
        bool
            True if download was successful, else False

        """
        if basedir is not None:
            self.basedir = basedir
        const.print_log.info(f'Downloading pyaerocom testdata into '
                             f'{self.basedir}')

        download_loc = self.basedir.joinpath(f'{self.TESTDATADIRNAME}.tar.gz')

        try:
            r = requests.get(self.URL_TESTDATA)
            with open(download_loc, 'wb') as f:
                f.write(r.content)

            with tarfile.open(download_loc, 'r:gz') as tar:
                tar.extractall(const.OUTPUTDIR)
                tar.close()
        except Exception:
            const.print_log.warning('Failed to download testdata. Traceback:\n{}'
                                    .format(format_exc()))
            return False
        finally:
            if download_loc.exists():
                os.remove(download_loc)
        return True

    def check_access(self, add_check_paths=None):
        """
        Method that checks if testdata can be accessed

        See also :func:`check_access_and_download_if_needed`.

        Parameters
        ----------
        add_check_paths : dict, optional
            additional paths used to validate testdata relative to testdata
            directory. Paths in :attr:`ADD_PATHS` will always be validated as
            they are required to be available in testdataset.

        Returns
        -------
        bool
            True if testdata is available and all relevant path locations
            could be validated, else False.

        """
        check_paths = {}
        check_paths.update(self.ADD_PATHS)
        if isinstance(add_check_paths, dict):
            check_paths.update(add_check_paths)

        datadir = self.testdatadir
        if not datadir.exists():
            return False

        for data_id, subdir in check_paths.items():
            if not datadir.joinpath(subdir).exists():
                return False
        return True

    def check_access_and_download_if_needed(self, add_check_paths=None):
        """
        Method that checks if testdata can be accessed and if not downloads it

        Parameters
        ----------
        add_check_paths : dict, optional
            additional paths used to validate testdata relative to testdata
            directory. Paths in :attr:`ADD_PATHS` will always be validated as
            they are required to be available in testdataset.

        Returns
        -------
        bool
            True if testdata is available and all relevant path locations
            could be validated, else False.

        """
        if not self.check_access(add_check_paths):
            try:
                if self.download() and self.check_access(add_check_paths):
                    return True
            except Exception as e:
                const.print_log.warning(f'Failed to access testdata: {e}')
            return False
        return True

    def init(self, add_check_paths=None):
        if not self.check_access_and_download_if_needed(add_check_paths):
            return False

        testdatadir = self.testdatadir
        for name, relpath in self.ADD_PATHS.items():
            ddir = str(testdatadir.joinpath(relpath))
            if name in self._UNGRIDDED_READERS:
                if name in const.OBSLOCS_UNGRIDDED and ddir == const.OBSLOCS_UNGRIDDED[name]:
                    const.print_log.info(f'dataset {name} is already registered')
                    continue
                reader = self._UNGRIDDED_READERS[name]
                try:
                    const.add_ungridded_obs(name, ddir,
                                            reader=reader,
                                            check_read=False)
                except Exception as e:
                    const.print_log.warning(
                        f'Failed to instantiate testdata since ungridded '
                        f'dataset {name} at {ddir} could not be registered: {e}')
                    return False
                const.print_log.info(
                    f'Adding ungridded dataset {name} located at {ddir}.'
                    f'Reader: {reader}')

            else:
                const.add_data_search_dir(ddir)
                const.print_log.info(f'Adding data search directory {ddir}.')
        return True

def initialise():
    td = AccessTestData()
    if td.init():
        const.print_log.info(f'pyaerocom-testdata is ready to be used. The data '
                             f'is available at {td.testdatadir}')
    else:
        const.print_log.warning('Failed to initiate pyaerocom-testdata')
if __name__ == '__main__':
    initialise()


################################################################
# read_aeronet_sunv2.py
#
# read Aeronet direct sun V2 data
#
# this file is part of the pyaerocom package
#
#################################################################
# Created 20171026 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2017 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA

import numpy as np
import pandas as pd
from collections import OrderedDict as od
import re

from pyaerocom import const
from pyaerocom.aux_var_helpers import calc_ang4487aer, calc_od550aer
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData

class ReadAeronetSunV2(ReadAeronetBase):
    """Interface for reading Aeronet direct sun version 2 Level 2.0 data

    Note
    ----
    Maintenance for AERONET version 2 data is deprecated since pyaerocom
    version 0.11.0. Please use version 3 data and associated pyaerocom readers.
    """
    #: Mask for identifying datafiles
    _FILEMASK = '*.lev20'

    #: version log of this class (for caching)
    __version__ = "0.16_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V2L2_AOD_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_SUN_V2L2_AOD_DAILY_NAME,
                          const.AERONET_SUN_V2L2_AOD_ALL_POINTS_NAME]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_SUN_V2L2_AOD_DAILY_NAME :   'daily'}

    #: default variables for read method
    DEFAULT_VARS = ['od550aer']

    #: value corresponding to invalid measurement
    NAN_VAL = -9999.

    #: Dictionary that specifies the index for each data column
    COL_INDEX = od(date           = 0,
                   time           = 1,
                   julien_day     = 2,
                   od1640aer      = 3,
                   od1020aer      = 4,
                   od870aer       = 5,
                   od675aer       = 6,
                   od667aer       = 7,
                   od555aer       = 8,
                   od551aer       = 9,
                   od532aer       = 10,
                   od531aer       = 11,
                   od500aer       = 12,
                   od440aer       = 15,
                   od380aer       = 17,
                   od340aer       = 18,
                   ang4487aer     = 37)

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {'od550aer'   :   ['od440aer',
                                      'od500aer',
                                      'ang4487aer'],
                    'ang4487aer_calc' :   ['od440aer',
                                           'od870aer']}

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {'od550aer'          :   calc_od550aer,
                'ang4487aer_calc'   :   calc_ang4487aer}

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = [k for k in COL_INDEX.keys()][3:]

    @property
    def col_index(self):
        """Dictionary that specifies the index for each data column

        Note
        ----
        Overload of method in base class
        """
        return self.COL_INDEX

    def read_file(self, filename, vars_to_retrieve=None,
                  vars_as_series=False):
        """Read Aeronet Sun V2 level 2 file

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results

        Example
        -------
        >>> import pyaerocom.io.read_aeronet_sunv2
        >>> obj = pyaerocom.io.read_aeronet_sunv2.ReadAeronetSunV2()
        >>> files = obj.get_file_list()
        >>> filedata = obj.read_file(files[0])
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id

        #create empty array for all variables that are supposed to be read
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.info("Reading file {}".format(filename))
        with open(filename, 'rt') as in_file:
            #added to output
            data_out.head_line = in_file.readline().strip()
            data_out.algorithm = in_file.readline().strip()
            c_dummy = in_file.readline()
            # re.split(r'=|\,',c_dummy)
            i_dummy = iter(re.split(r'=|\,', c_dummy.rstrip()))
            dict_loc = dict(zip(i_dummy, i_dummy))

            data_out['latitude'] = float(dict_loc['lat'])
            data_out['longitude'] = float(dict_loc['long'])
            data_out['altitude'] = float(dict_loc['elev'])
            data_out['station_name'] = dict_loc['Location']
            data_out['PI'] = dict_loc['PI']
            data_out['ts_type'] = self.TS_TYPE
            data_out['instrument_name'] = self.INSTRUMENT_NAME

            c_dummy = in_file.readline()

            #data_out.data_header =
            in_file.readline().strip()

            for line in in_file:
                # process line
                dummy_arr = line.split(self.COL_DELIM)

                day, month, year = dummy_arr[self.col_index['date']].split(':')

                datestring = '-'.join([year, month, day])
                datestring = 'T'.join([datestring, dummy_arr[self.col_index['time']]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                #datestring = '+'.join([datestring, '00:00'])

                data_out['dtime'].append(np.datetime64(datestring))

                for var in vars_to_read:
                    val = float(dummy_arr[self.col_index[var]])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)
        data_out['dtime'] = np.asarray(data_out['dtime'])
        for var in vars_to_read:
            data_out[var] = np.asarray(data_out[var])

        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        # TODO: reconsider to skip conversion to Series
        # convert  the vars in vars_to_retrieve to pandas time series
        # and delete the other ones
        if vars_as_series:
            for var in (vars_to_read + vars_to_compute):
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var],
                                              index=data_out['dtime'])
                else:
                    del data_out[var]

        return data_out

if __name__=="__main__":
    import pyaerocom as pya
    pya.change_verbosity('critical')
    data = {}
    failed = []
    for name in ReadAeronetSunV2.SUPPORTED_DATASETS:
        print('reading {}'.format(name))
        reader = ReadAeronetSunV2(name)
        try:
            data[name] = reader.read_first_file()
        except Exception as e:
            failed.append(repr(e))

    for name, sdata in data.items():
        print(name)
        print(sdata)

    read = ReadAeronetSunV2()

    read.verbosity_level = 'critical'
    first_ten = read.read(last_file=10)

    data0 = read.read_first_file(vars_as_series=True)
    data = read.read_first_file()
    print(data)

    stats = first_ten.to_station_data_all()

    read = ReadAeronetSunV2(const.AERONET_SUN_V2L2_AOD_ALL_POINTS_NAME)
    data = read.read_first_file(vars_to_retrieve=['ang4487aer',
                                                  'ang4487aer_calc'])
    print(data)



################################################################
# read_aeronet_invv3.py
#
# read Aeronet inversion V3 data
#
# this file is part of the pyaerocom package
#
#################################################################
# Created 20180802 by Jonas Gliss for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2018 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA

import os

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.aux_var_helpers import calc_od550aer, calc_abs550aer
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData

class ReadAeronetInvV3(ReadAeronetBase):
    """Interface for reading Aeronet inversion V3 Level 1.5 and 2.0 data

    Parameters
    ----------
    data_id
        string specifying either of the supported datasets that are defined
        in ``SUPPORTED_DATASETS``
    """
    #: Mask for identifying datafiles
    _FILEMASK = '*.all'

    #: version log of this class (for caching)
    __version__ = "0.03_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_INV_V3L2_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_INV_V3L2_DAILY_NAME,
                          const.AERONET_INV_V3L15_DAILY_NAME]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_INV_V3L15_DAILY_NAME   :   'daily',
                const.AERONET_INV_V3L2_DAILY_NAME    :   'daily'}

    #: default variables for read method
    DEFAULT_VARS = ['abs550aer',
                    'od550aer']

    #: value corresponding to invalid measurement
    NAN_VAL = -999.

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {'abs550aer'     :   ['abs440aer',
                                         'angabs4487aer'],
                    'od550aer'     :   ['od440aer',
                                        'ang4487aer']}

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {'abs550aer'     :   calc_abs550aer,
                'od550aer'      :   calc_od550aer}

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE['abs440aer'] = 'Absorption_AOD[440nm]'
    VAR_NAMES_FILE['angabs4487aer'] = 'Absorption_Angstrom_Exponent_440-870nm'
    VAR_NAMES_FILE['od440aer'] = 'AOD_Extinction-Total[440nm]'
    VAR_NAMES_FILE['ang4487aer'] = 'Extinction_Angstrom_Exponent_440-870nm-Total'

    #: dictionary specifying the file column names (values) for each
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}
    #META_NAMES_FILE['data_quality_level'] = 'DATA_TYPE'
    META_NAMES_FILE['date'] = 'Date(dd:mm:yyyy)'
    META_NAMES_FILE['time'] = 'Time(hh:mm:ss)'
    META_NAMES_FILE['day_of_year'] = 'Day_of_Year(fraction)'
    META_NAMES_FILE['station_name'] = 'AERONET_Site'
    META_NAMES_FILE['latitude'] = 'Latitude(Degrees)'
    META_NAMES_FILE['longitude'] = 'Longitude(Degrees)'
    META_NAMES_FILE['altitude'] = 'Elevation(m)'

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE.keys())

    def read_file(self, filename, vars_to_retrieve=None,
                  vars_as_series=False):
        """Read Aeronet file containing results from v2 inversion algorithm

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : list
            list of str with variable names to read
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results

        Example
        -------
        >>> import pyaerocom.io as pio
        >>> obj = pio.read_aeronet_invv2.ReadAeronetInvV2()
        >>> files = obj.get_file_list()
        >>> filedata = obj.read_file(files[0])
        """
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()

        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.debug("Reading file {}".format(filename))

        with open(filename, 'rt') as in_file:

            data_out['dataset_info'] = in_file.readline().strip()
            self.logger.debug('Skipping line: {}'.format(in_file.readline()))
            data_out['algorithm_info'] = in_file.readline().strip()

            self.logger.debug('Skipping line: {}'.format(in_file.readline()))

            c_dummy = in_file.readline().strip().split(',')
            data_out['freq_info'] = c_dummy[0].strip()

            pi_info = c_dummy[1].strip().split(';')
            # re.split(r'=|\,',c_dummy)

            data_out['PI'] = pi_info[0].split('PI=')[1].strip()
            data_out['PI_email'] = pi_info[1].split('PI Email=')[1].strip()
            data_out['ts_type'] = self.TS_TYPE

            #skip next two lines
            self.logger.debug('Skipping line:\n{}'.format(in_file.readline()))
            #self.logger.info('Skipping line:\n{}'.format(in_file.readline()))

            col_index_str = in_file.readline()
            if col_index_str != self._last_col_index_str:
                self.logger.debug("Header has changed, reloading col_index map")
                self._update_col_index(col_index_str)
            col_index = self.col_index

            # dependent on the station, some of the required input variables
            # may not be provided in the data file. These will be ignored
            # in the following list that iterates over all data rows and will
            # be filled below, with vectors containing NaNs after the file
            # reading loop
            vars_available = {}
            for var in vars_to_read:
                if var in col_index:
                    vars_available[var] = col_index[var]
                else:
                    self.logger.warning("Variable {} not available in file {}"
                                        .format(var, os.path.basename(filename)))

            for line in in_file:
                # process line
                dummy_arr = line.strip().split(self.COL_DELIM)

                # This uses the numpy datestring64 functions that i.e. also
                # support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string
                day, month, year = dummy_arr[col_index['date']].split(':')
                datestring = '-'.join([year, month, day])
                datestring = 'T'.join([datestring, dummy_arr[col_index['time']]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                #datestring = '+'.join([datestring, '00:00'])
                data_out['dtime'].append(np.datetime64(datestring))

                for var in self.META_NAMES_FILE:
                    val = dummy_arr[col_index[var]]
                    try:
                        # e.g. lon, lat, altitude
                        val = float(val)
                    except Exception:
                        pass
                    data_out[var].append(val)

                # copy the data fields that are available (rest will be filled
                # below)
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)

        # convert all lists to numpy arrays
        data_out['dtime'] = np.asarray(data_out['dtime'])

        for item in self.META_NAMES_FILE:
            data_out[item] = np.asarray(data_out[item])

        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out['dtime'])) * np.nan
            data_out[var] = array

        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        if vars_as_series:
            for var in (vars_to_read + vars_to_compute):
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var],
                                              index=data_out['dtime'])
                else:
                    del data_out[var]

        return data_out

if __name__=="__main__":
    from pyaerocom import change_verbosity
    change_verbosity('critical')
    read = ReadAeronetInvV3()
    read15 = ReadAeronetInvV3(const.AERONET_INV_V3L15_DAILY_NAME)
    read.verbosity_level = 'info'

    data = read.read_first_file()
    data15 = read15.read_first_file()
    print(data)

    udat = read.read(last_file=10)



import fnmatch
import os
import re
from collections import OrderedDict as od

import numpy as np
import xarray

from pyaerocom import const
from pyaerocom.exceptions import DataUnitError
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.units_helpers import get_unit_conversion_fac
from pyaerocom.variable import Variable
from pyaerocom.vertical_profile import VerticalProfile

class ReadEarlinet(ReadUngriddedBase):
    """Interface for reading of EARLINET data"""

    #: Mask for identifying datafiles
    _FILEMASK = '*.*'

    #: version log of this class (for caching)
    __version__ = "0.15_" + ReadUngriddedBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.EARLINET_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.EARLINET_NAME]

    #: default variables for read method
    DEFAULT_VARS = ['ec532aer']

    #: all data values that exceed this number will be set to NaN on read. This
    #: is because iris, xarray, etc. assign a FILL VALUE of the order of e36
    #: to missing data in the netcdf files
    _MAX_VAL_NAN = 1e6

    #: variable name of altitude in files
    ALTITUDE_ID = 'Altitude'

    #: temporal resolution
    TS_TYPE = 'native'

    #: dictionary specifying the file search patterns for each variable
    VAR_PATTERNS_FILE = {'ec532aer'     : '*.e532',
                         'ec355aer'     : '*.e355',
                         'bsc532aer'    : '*.b532',
                         'bsc355aer'    : '*.b355',
                         'bsc1064aer'   : '*.b1064',
                         'zdust'        : '*.e*'}

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {'ec532aer'        : 'Extinction',
                      'ec355aer'        : 'Extinction',
                      'ec1064aer'       : 'Extinction',
                      'bsc532aer'       : 'Backscatter',
                      'bsc355aer'       : 'Backscatter',
                      'bsc1064aer'      : 'Backscatter',
                      'zdust'           : 'DustLayerHeight'}

    #: metadata names that are supposed to be imported
    META_NAMES_FILE = od(location           = 'Location',
                         start_date         = 'StartDate',
                         start_utc          = 'StartTime_UT',
                         stop_utc           = 'StopTime_UT',
                         longitude          = 'Longitude_degrees_east',
                         latitude           = 'Latitude_degrees_north',
                         wavelength_emis    = 'EmissionWavelength_nm',
                         wavelength_det     = 'DetectionWavelength_nm',
                         res_raw_m          = 'ResolutionRaw_meter',
                         zenith_ang_deg     = 'ZenithAngle_degrees',
                         instrument_name    = 'System',
                         comments           = 'Comments',
                         shots_avg          = 'ShotsAveraged',
                         detection_mode     = 'DetectionMode',
                         res_eval           = 'ResolutionEvaluated',
                         input_params       = 'InputParameters',
                         altitude           = 'Altitude_meter_asl',
                         eval_method        = 'EvaluationMethod')

    #: metadata keys that are needed for reading (must be values in
    #: :attr:`META_NAMES_FILE`)
    META_NEEDED = ['Location', 'StartDate', 'StartTime_UT', 'StopTime_UT',
                   'Longitude_degrees_east', 'Latitude_degrees_north',
                   'Altitude_meter_asl']

    #: Metadata keys from :attr:`META_NAMES_FILE` that are additional to
    #: standard keys defined in :class:`StationMetaData` and that are supposed
    #: to be inserted into :class:`UngriddedData` object created in :func:`read`
    KEEP_ADD_META = ['location', 'wavelength_emis', 'wavelength_det',
                     'res_raw_m', 'zenith_ang_deg', 'comments', 'shots_avg',
                     'detection_mode', 'res_eval', 'input_params',
                     'eval_method']

    #: Attribute access names for unit reading of variable data
    VAR_UNIT_NAMES = dict(
        Extinction      = ['ExtinctionUnits', 'units'],
        Backscatter     = ['BackscatterUnits', 'units'],
        DustLayerHeight = ['units'],
        Altitude        = 'units')
    #: Variable names of uncertainty data
    ERR_VARNAMES = dict(
        ec532aer = 'ErrorExtinction',
        ec355aer = 'ErrorExtinction'
    )

    #: If true, the uncertainties are also read (where available, cf. ERR_VARNAMES)
    READ_ERR = True

    PROVIDES_VARIABLES = list(VAR_PATTERNS_FILE)

    EXCLUDE_CASES = ['cirrus.txt']

    def __init__(self, data_id=None, data_dir=None):
        # initiate base class
        super(ReadEarlinet, self).__init__(data_id=data_id,
                                           data_dir=data_dir)
        # make sure everything is properly set up
        if not all([x in self.VAR_PATTERNS_FILE for x in
                    self.PROVIDES_VARIABLES]): # pragma: no cover
            raise AttributeError("Please specify file search masks in "
                                 "header dict VAR_PATTERNS_FILE for each "
                                 "variable defined in PROVIDES_VARIABLES")
        elif not all([x in self.VAR_NAMES_FILE for x in
                      self.PROVIDES_VARIABLES]): # pragma: no cover
            raise AttributeError("Please specify file search masks in "
                                 "header dict VAR_NAMES_FILE for each "
                                 "variable defined in PROVIDES_VARIABLES")
        #: private dictionary containing loaded Variable instances,
        self._var_info = {}

        #: files that are supposed to be excluded from reading
        self.exclude_files = []

        #: files that were actually excluded from reading
        self.excluded_files = []

    def read_file(self, filename, vars_to_retrieve=None, read_err=None,
                  remove_outliers=True):
        """Read EARLINET file and return it as instance of :class:`StationData`

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        read_err : bool
            if True, uncertainty data is also read (where available).
        remove_outliers : bool
            if True, outliers are removed for each variable using the
            `minimum` and `maximum` attributes for that variable (accessed
            via pyaerocom.const.VARS[var_name]).

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if read_err is None: #use default setting
            read_err = self.READ_ERR
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        _vars = []
        for var in vars_to_retrieve:
            if var in self.VAR_PATTERNS_FILE: #make sure to only read what is supported by this file
                if fnmatch.fnmatch(filename, self.VAR_PATTERNS_FILE[var]):
                    _vars.append(var)
            elif var in self.AUX_REQUIRES:
                _vars.append(var)
            else:
                raise ValueError(f'{var} is not supported')

        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(_vars)

        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out['station_id'] = filename.split('/')[-2]
        data_out['data_id'] = self.data_id
        data_out['ts_type'] = self.TS_TYPE

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            if not var in self._var_info:
                self._var_info[var] = Variable(var)
        var_info = self._var_info

        # Iterate over the lines of the file
        self.logger.debug("Reading file {}".format(filename))

        data_in = xarray.open_dataset(filename)

        for k, v in self.META_NAMES_FILE.items():
            if v in self.META_NEEDED:
                _meta = data_in.attrs[v]
            else:
                try:
                    _meta = data_in.attrs[v]
                except Exception: # pragma: no cover
                    _meta = None
            data_out[k] = _meta

        data_out['station_name'] = re.split('\s|,', data_out['location'])[0].strip()

        str_dummy = str(data_in.StartDate)
        year, month, day = str_dummy[0:4], str_dummy[4:6], str_dummy[6:8]

        str_dummy = str(data_in.StartTime_UT).zfill(6)
        hours, minutes, seconds = str_dummy[0:2], str_dummy[2:4], str_dummy[4:6]

        datestring = '-'.join([year, month, day])
        startstring = 'T'.join([datestring, ':'.join([hours, minutes, seconds])])

        dtime = np.datetime64(startstring)

        str_dummy = str(data_in.StopTime_UT).zfill(6)
        hours, minutes, seconds = str_dummy[0:2], str_dummy[2:4], str_dummy[4:6]

        stopstring = 'T'.join([datestring, ':'.join([hours, minutes, seconds])])

        stop = np.datetime64(stopstring)

        # in case measurement goes over midnight into a new day
        if stop < dtime:
            stop = stop + np.timedelta64(1, '[D]')

        data_out['dtime'] = [dtime]
        data_out['stopdtime'] = [stop]
        data_out['has_zdust'] = False

        for var in vars_to_read:
            data_out['var_info'][var] = od()
            err_read = False
            unit_ok = False
            outliers_removed = False

            netcdf_var_name = self.VAR_NAMES_FILE[var]
            # check if the desired variable is in the file
            if netcdf_var_name not in data_in.variables:
                self.logger.warning("Variable {} not found in file {}".format(var, filename))
                continue

            info = var_info[var]
            # xarray.DataArray
            arr = data_in.variables[netcdf_var_name]
            # the actual data as numpy array (or float if 0-D data, e.g. zdust)
            val = np.float64(arr)

            # CONVERT UNIT
            unit = None

            unames = self.VAR_UNIT_NAMES[netcdf_var_name]
            for u in unames:
                if u in arr.attrs:
                    unit = arr.attrs[u]
            if unit is None:
                raise DataUnitError('Unit of {} could not be accessed in file '
                                    '{}'.format(var, filename))
            unit_fac = None
            try:
                to_unit = self._var_info[var].units
                unit_fac = get_unit_conversion_fac(unit, to_unit)
                val *= unit_fac
                unit = to_unit
                unit_ok = True
            except Exception as e:
                const.print_log.warning('Failed to convert unit of {} in file '
                                        '{} (Earlinet): Error: {}'
                                        .format(var, filename, repr(e)))

            # import errors if applicable
            err = np.nan
            if read_err and var in self.ERR_VARNAMES:
                err_name = self.ERR_VARNAMES[var]
                if err_name in data_in.variables:
                    err = np.float64(data_in.variables[err_name])
                    if unit_ok:
                        err *= unit_fac
                    err_read = True

            # 1D variable
            if var == 'zdust':
                if not val.ndim == 0:
                    raise ValueError('Fatal: dust layer height data must be '
                                     'single value')

                if unit_ok and info.minimum < val < info.maximum:
                    const.print_log.warning('zdust value {} out of range, '
                                            'setting to NaN'.format(val))
                    val = np.nan

                if np.isnan(val):
                    self.logger.warning("Invalid value of variable zdust "
                                        "in file {}. Skipping...!".format(filename))
                    continue

                data_out['has_zdust'] = True
                data_out[var] = val

            else:
                if not val.ndim == 1:
                    raise ValueError('Extinction data must be one dimensional')
                elif len(val) == 0:
                    continue # no data
                # Remove NaN equivalent values
                val[val>self._MAX_VAL_NAN] = np.nan

                wvlg = var_info[var].wavelength_nm
                wvlg_str = self.META_NAMES_FILE['wavelength_emis']

                if not wvlg == data_in.attrs[wvlg_str]:
                    self.logger.info('No wavelength match')
                    continue

                alt_id = self.ALTITUDE_ID
                alt_data = data_in.variables[alt_id]

                alt_vals = np.float64(alt_data)
                alt_unit = alt_data.attrs[self.VAR_UNIT_NAMES[alt_id]]
                to_alt_unit = const.VARS['alt'].units
                if not alt_unit == to_alt_unit:
                    try:
                        alt_unit_fac = get_unit_conversion_fac(alt_unit,
                                                               to_alt_unit)
                        alt_vals *= alt_unit_fac
                        alt_unit = to_alt_unit
                    except Exception as e:
                        self.logger.warning('Failed to convert unit: {}'.format(repr(e)))

                # remove outliers from data, if applicable
                if remove_outliers and unit_ok:
                    # REMOVE OUTLIERS
                    outlier_mask = np.logical_or(val < info.minimum,
                                                 val > info.maximum)
                    val[outlier_mask] = np.nan

                    if err_read:
                        err[outlier_mask] = np.nan
                    outliers_removed = True
                # remove outliers from errors if applicable
                if err_read:
                    err[err > self._MAX_VAL_NAN] = np.nan

                # create instance of ProfileData
                profile = VerticalProfile(data=val,
                                          altitude=alt_vals,
                                          dtime=dtime,
                                          var_name=var,
                                          data_err=err,
                                          var_unit=unit,
                                          altitude_unit=alt_unit)
                # Write everything into profile
                data_out[var] = profile

            data_out['var_info'][var].update(unit_ok=unit_ok,
                                             err_read=err_read,
                                             outliers_removed=outliers_removed)

        return (data_out)

    def read(self, vars_to_retrieve=None, files=None, first_file=None,
             last_file=None, read_err=None, remove_outliers=True,
             pattern=None):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used
        read_err : bool
            if True, uncertainty data is also read (where available). If
            unspecified (None), then the default is used (cf. :attr:`READ_ERR`)
         pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)

        Returns
        -------
        UngriddedData
            data object
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        if read_err is None:
            read_err = self.READ_ERR

        if files is None:
            if len(self.files) == 0:
                self.get_file_list(vars_to_retrieve, pattern=pattern)
            files = self.files

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        files = files[first_file:last_file]

        self.read_failed = []

        data_obj = UngriddedData()
        col_idx = data_obj.index
        meta_key = -1.0
        idx = 0

        #assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        #last_station_id = ''
        num_files = len(files)

        disp_each = int(num_files*0.1)
        if disp_each < 1:
            disp_each = 1

        VAR_IDX = -1
        for i, _file in enumerate(files):
            if i%disp_each == 0:
                print("Reading file {} of {} ({})".format(i+1,
                                 num_files, type(self).__name__))
            try:
                stat = self.read_file(_file,
                                      vars_to_retrieve=vars_to_retrieve,
                                      read_err=read_err,
                                      remove_outliers=remove_outliers)
                if not any([var in stat.vars_available for var in
                            vars_to_retrieve]):
                    self.logger.info("Station {} contains none of the desired "
                                     "variables. Skipping station..."
                                     .format(stat.station_name))
                    continue
                #if last_station_id != station_id:
                meta_key += 1
                # Fill the metatdata dict
                # the location in the data set is time step dependant!
                # use the lat location here since we have to choose one location
                # in the time series plot
                metadata[meta_key] = od()
                metadata[meta_key].update(stat.get_meta())
                for add_meta in self.KEEP_ADD_META:
                    if add_meta in stat:
                        metadata[meta_key][add_meta] = stat[add_meta]
                #metadata[meta_key]['station_id'] = station_id

                metadata[meta_key]['data_revision'] = self.data_revision
                metadata[meta_key]['variables'] = []
                metadata[meta_key]['var_info'] = od()
                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = od()
                    #last_station_id = station_id

                # Is floating point single value
                time = stat.dtime[0]
                for var in stat.vars_available:
                    if not var in data_obj.var_idx:
                        VAR_IDX +=1
                        data_obj.var_idx[var] = VAR_IDX

                    var_idx = data_obj.var_idx[var]

                    val = stat[var]
                    metadata[meta_key]['var_info'][var] = vi = od()
                    if isinstance(val, VerticalProfile):
                        altitude = val.altitude
                        data = val.data
                        add = len(data)
                        err = val.data_err
                        metadata[meta_key]['var_info']['altitude'] = via = od()

                        vi.update(val.var_info[var])
                        via.update(val.var_info['altitude'])
                    else:
                        add = 1
                        altitude = np.nan
                        data = val
                        if var in stat.data_err:
                            err = stat.err[var]
                        else:
                            err = np.nan
                    vi.update(stat.var_info[var])
                    stop = idx + add
                    #check if size of data object needs to be extended
                    if stop >= data_obj._ROWNO:
                        #if totnum < data_obj._CHUNKSIZE, then the latter is used
                        data_obj.add_chunk(add)

                    #write common meta info for this station
                    data_obj._data[idx:stop,
                                   col_idx['latitude']] = stat['latitude']
                    data_obj._data[idx:stop,
                                   col_idx['longitude']] = stat['longitude']
                    data_obj._data[idx:stop,
                                   col_idx['altitude']] = stat['altitude']
                    data_obj._data[idx:stop,
                                   col_idx['meta']] = meta_key

                    # write data to data object
                    data_obj._data[idx:stop, col_idx['time']] = time
                    data_obj._data[idx:stop, col_idx['stoptime']] = stat.stopdtime[0]
                    data_obj._data[idx:stop, col_idx['data']] = data
                    data_obj._data[idx:stop, col_idx['dataaltitude']] = altitude
                    data_obj._data[idx:stop, col_idx['varidx']] = var_idx

                    if read_err:
                        data_obj._data[idx:stop, col_idx['dataerr']] = err

                    if not var in meta_idx[meta_key]:
                        meta_idx[meta_key][var] = []
                    meta_idx[meta_key][var].extend(list(range(idx, stop)))

                    if not var in metadata[meta_key]['variables']:
                        metadata[meta_key]['variables'].append(var)

                    idx += add

            except Exception as e:
                self.read_failed.append(_file)
                self.logger.exception('Failed to read file {} (ERR: {})'
                                      .format(os.path.basename(_file),
                                              repr(e)))

        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]

        self.data = data_obj
        return data_obj


    def _get_exclude_filelist(self): # pragma: no cover
        """Get list of filenames that are supposed to be ignored"""
        exclude = []
        import glob
        files = glob.glob('{}/EXCLUDE/*.txt'.format(self.data_dir))
        for i, file in enumerate(files):
            if not os.path.basename(file) in self.EXCLUDE_CASES:
                continue
            count = 0
            num = None
            indata = False
            with open(file) as f:
                for line in f:
                    if indata:
                        exclude.append(line.strip())
                        count += 1
                    elif 'Number of' in line:
                        num = int(line.split(':')[1].strip())
                        indata = True

            if not count == num:
                raise Exception
        self.exclude_files = list(dict.fromkeys(exclude))
        return self.exclude_files


    def get_file_list(self, vars_to_retrieve=None, pattern=None):
        """Perform recusive file search for all input variables

        Note
        ----
        Overloaded implementation of base class, since for Earlinet, the
        paths are variable dependent

        Parameters
        ----------
        vars_to_retrieve : list
            list of variables to retrieve
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing file paths
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        exclude = self._get_exclude_filelist()
        const.print_log.info('Fetching EARLINET data files. '
                             'This might take a while...')
        patterns = []
        for var in vars_to_retrieve:
            if not var in self.VAR_PATTERNS_FILE:
                from pyaerocom.exceptions import VarNotAvailableError
                raise VarNotAvailableError('Input variable {} is not supported'
                                           .format(var))

            _pattern = self.VAR_PATTERNS_FILE[var]
            if pattern is not None:
                if '.' in pattern:
                    raise NotImplementedError('filetype delimiter . not '
                                              'supported')
                spl = _pattern.split('.')
                if not '*' in spl[0]:
                    raise AttributeError('Invalid file pattern: {}'
                                         .format(_pattern))
                spl[0] = spl[0].replace('*', pattern)
                _pattern = '.'.join(spl)

            patterns.append(_pattern)

        matches = []
        for root, dirnames, files in os.walk(self.data_dir):
            paths = [os.path.join(root, f) for f in files]
            for _pattern in patterns:
                for path in paths:
                    file = os.path.basename(path)
                    if not fnmatch.fnmatch(file, _pattern):
                        continue
                    elif file in exclude:
                        self.excluded_files.append(path)
                    else:
                        matches.append(path)
        self.files = files = list(dict.fromkeys(matches))
        return files


"""
Methods to convert different standards of vertical coordinates

For details see here:

http://cfconventions.org/Data/cf-conventions/cf-conventions-1.0/build/apd.html

Note
----
UNDER DEVELOPMENT -> NOT READY YET

"""

from pyaerocom import const
from pyaerocom.exceptions import (CoordinateNameError,
                                  VariableNotFoundError,
                                  VariableDefinitionError,
                                  DataDimensionError)

def atmosphere_sigma_coordinate_to_pressure(sigma, ps, ptop):
    """Convert atmosphere sigma coordinate to pressure in Pa

    Note
    ----
    This formula only works at one lon lat coordinate and at one instant in
    time.

    **Formula**:

    .. math::

        p(k) = p_{top} + \\sigma(k) \\cdot (p_{surface} - p_{top})

    Parameters
    ----------
    sigma : ndarray or float
        sigma coordinate (1D) array
    ps : float
        surface pressure
    ptop : float
        ToA pressure

    Returns
    -------
    ndarray or float
        computed pressure levels in Pa (standard_name=air_pressure)
    """
    if not isinstance(ptop, float):
        try:
            ptop = float(ptop)
        except Exception as e:
            raise ValueError('Invalid input for ptop. Need floating point\n'
                             'Error: {}'.format(repr(e)))
    return ptop + sigma * (ps - ptop)

def atmosphere_hybrid_sigma_pressure_coordinate_to_pressure(a, b, ps, p0=None):
    """Convert atmosphere_hybrid_sigma_pressure_coordinate to  pressure in Pa

    **Formula**:

    Either

    .. math::

        p(k) = a(k) \\cdot p_0 + b(k) \\cdot p_{surface}

    or

    .. math::

        p(k) = ap(k) + b(k) \\cdot p_{surface}

    Parameters
    ----------
    a : ndarray
        sigma level values (a(k) in formula 1, and ap(k) in formula 2)
    b : ndarray
        dimensionless fraction per level (must be same length as a)
    ps : float
        surface pressure
    p0 :
        reference pressure (only relevant for alternative formula 1)

    Returns
    -------
    ndarray
        computed pressure levels in Pa (standard_name=air_pressure)

    """
    if not len(a) == len(b):
        raise ValueError('Invalid input: a and b must have the same length')
    if p0 is None:
        return a + b*ps
    return a*p0 + b*ps

def geopotentialheight2altitude(geopotential_height):
    """Convert geopotential height in m to altitude in m

    Note
    ----
    This is a dummy function that returns the input, as the conversion is not
    yet implemented.

    Parameters
    ----------
    geopotential_height
        input geopotential height values in m

    Returns
    -------
    Computed altitude levels
    """

    const.print_log.warning('Conversion method of geopotential height to '
                            'altitude is not yet implemented and returns the '
                            'input values. The introduced error is small at '
                            'tropospheric altitudes')
    return geopotential_height

def pressure2altitude(p, *args, **kwargs):
    """General formula to convert atm. pressure to altitude

    Wrapper method for :func:`geonum.atmosphere.pressure2altitude`

    **Formula:**

    .. math::

        h = h_{ref} + \\frac{T_{ref}}{L} \\left(\\exp\\left[-\\frac{\\ln\\left(\\frac{p}{p_{ref}} \\right)}{\\beta}\\right] - 1\\right) \\quad [m]

    where:

        - :math:`$h_{ref}$` is a reference altitude
        - :math:`$T_{ref}$` is a reference temperature
        - :math:`$L$` is the atmospheric lapse-rate (cf. :attr:`L_STD_ATM`, \
                                                     :attr:`L_DRY_AIR`)
        - :math:`$p$` is the pressure (cf. :func:`pressure`)
        - :math:`$p_{ref}$` is a reference pressure
        - :math:`$\\beta$` is computed using :func:`beta_exp`

    Parameters
    ----------
    p
        pressure in Pa
    *args
        additional non-keyword args passed to
        :func:`geonum.atmosphere.pressure2altitude`
    **kwargs
        additional keyword args passed to
        :func:`geonum.atmosphere.pressure2altitude`

    Returns
    -------
    altitudes in m corresponding to input pressure levels in defined atmosphere
    """
    if not const.GEONUM_AVAILABLE:
        raise ModuleNotFoundError('Feature disabled: need geonum library')
    from geonum import atmosphere as atm
    return atm.pressure2altitude(p, *args, **kwargs)

def is_supported(standard_name):
    """Checks if input coordinate standard name is supported by pyaerocom

    Parameters
    ----------
    standard_name : str
        standard name of vertical coordinate

    Returns
    -------
    bool
        True, if this coordinate is supported, else False
    """
    return True if standard_name in VerticalCoordinate.REGISTERED else False

class VerticalCoordinate(object):

    NAMES_SUPPORTED = {
    'altitude'                                      : 'z',
    'air_pressure'                                  : 'pres',
    'geopotential_height'                           : 'gph',
    'atmosphere_sigma_coordinate'                   : 'asc',
    'atmosphere_hybrid_sigma_pressure_coordinate'   : 'ahspc',
    }

    STANDARD_NAMES = dict(map(reversed, NAMES_SUPPORTED.items()))

    NAMES_NOT_SUPPORTED = ['model_level_number']

    #: registered names
    REGISTERED = list(NAMES_SUPPORTED) + NAMES_NOT_SUPPORTED

    CONVERSION_METHODS = {
    'asc'   :   atmosphere_sigma_coordinate_to_pressure,
    'ahspc' :   atmosphere_hybrid_sigma_pressure_coordinate_to_pressure,
    'gph'   :   geopotentialheight2altitude
    }

    CONVERSION_REQUIRES = {

    'asc'   :   ['sigma', 'ps', 'ptop'],
    'ahspc' :   ['a', 'b', 'ps', 'p0'],
    'gph'   :   []
    }

    FUNS_YIELD = {
    'asc'   :   'air_pressure',
    'ahspc' :   'air_pressure',
    'gph'   :   'altitude'
    }

    _LEV_INCREASES_WITH_ALT = dict(
        atmosphere_sigma_coordinate=False,
        atmosphere_hybrid_sigma_pressure_coordinate=False)

    def __init__(self, name=None):
        self.var_name = None
        self.standard_name = None
        self._eval_input(name)

    def _eval_input(self, name):
        if name in self.NAMES_SUPPORTED:
            self.var_name = self.NAMES_SUPPORTED[name]
            self.standard_name = name
        elif name in self.STANDARD_NAMES:
            self.var_name = name
            self.standard_name = self.STANDARD_NAMES[name]
        elif name in self.NAMES_NOT_SUPPORTED:
            self.standard_name = name
        else:
            raise ValueError('Invalid name for vertical coordinate: {}'
                             .format(name))

    @property
    def fun(self):
        """Function used to convert levels into pressure"""
        return self.CONVERSION_METHODS[self.var_name]

    @property
    def conversion_requires(self):
        """Valid argument names for :func:`fun`"""
        return self.CONVERSION_REQUIRES[self.var_name]

    @property
    def conversion_supported(self):
        """Boolean specifying whether a conversion scheme is provided"""
        return True if self.standard_name in self.NAMES_SUPPORTED else False

    @property
    def lev_increases_with_alt(self):
        """Boolean specifying whether coordinate levels increase with altitude

        Returns
        -------
        True
        """
        if not self.var_name in self._LEV_INCREASES_WITH_ALT:
            if self.standard_name in self._LEV_INCREASES_WITH_ALT:
                return self._LEV_INCREASES_WITH_ALT[self.standard_name]
            raise ValueError('Failed to access information '
                             'lev_increases_with_alt for vertical coordinate {}'
                             .format(self.var_name))
        return self._LEV_INCREASES_WITH_ALT[self.var_name]

    def calc_pressure(self, lev, **kwargs):
        """Compute pressure levels for input vertical coordinate

        Parameters
        ----------
        vals
            level values that are supposed to be converted into pressure
        **kwargs
            additional  keyword args required for computation of pressure
            levels (cf. :attr:`CONVERSION_METHODS` and corresponding inputs for method
            available)

        Returns
        -------
        ndarray
            pressure levels in Pa
        """

        if not self.var_name in self.NAMES_SUPPORTED:

            raise CoordinateNameError('Variable {} cannot be converted to '
                                      'pressure levels. Conversion is only '
                                      'possible for supported variables:\n{}'
                                      .format(self.var_name,
                                              self.vars_supported_str))

        coord_values = kwargs.pop(self.var_name)
        return self.fun(coord_values, **kwargs)

    @property
    def vars_supported_str(self):
        from pyaerocom._lowlevel_helpers import dict_to_str
        return dict_to_str(self.NAMES_SUPPORTED)

    def pressure2altitude(self, p, **kwargs):
        """Convert pressure to altitude

        Wrapper for method

        Parameters
        ----------

        Returns
        -------
        """
        raise NotImplementedError('Coming soon...')
        #pressure2altitude(p, **kwargs)
        return

class AltitudeAccess(object):

    #: Additional variable names (in AEROCOM convention) that are used
    #: to search for additional files that can be used to access or compute
    #: the altitude levels at each grid point
    ADD_FILE_VARS = ['z', 'z3d', 'pres', 'deltaz3d']

    #: Additional variables that are required to compute altitude levels
    ADD_FILE_REQ = {'deltaz3d' : ['ps']}

    ADD_FILE_OPT = {'pres'  :   ['temp']}

    def __init__(self, gridded_data):
        from pyaerocom import logger
        from pyaerocom.griddeddata import GriddedData
        if not isinstance(gridded_data, GriddedData):
            raise ValueError('Invalid input: require instance of GriddedData '
                             'class')
        if not gridded_data.has_latlon_dims:
            raise NotImplementedError('Altitude access requires latitude and '
                                      'longitude dimensions to be available '
                                      'in  input GriddedData: {}'
                                      .format(gridded_data.short_str()))
        self.data_obj = gridded_data
        self._subset1d = None
        self._checked_and_failed = []
        self._has_access = False
        self.logger = logger

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __getitem__(self, key):
        return self.__dict__[key]

    def __contains__(self, key):
        return True if key in self.__dict__ else False

    @property
    def has_access(self):
        """Boolean specifying whether altitudes can be accessed

        Note
        ----
        Performs access check using :func:`check_altitude_access` if access
        flag is False
        """
        if not self._has_access:
            self._has_access = self.check_altitude_access()
        return self._has_access

    @property
    def coord_list(self):
        """List of AeroCom coordinate names for altitude access"""
        l = self.ADD_FILE_VARS + list(VerticalCoordinate.NAMES_SUPPORTED.values())
        return list(dict.fromkeys(l))

    def extract_1D_subset_from_data(self, **coord_info):
        """Extract 1D subset containing only vertical coordinate dimension

        Note
        ----
        So far this Works only for 4D or 3D data that contains latitude and
        longitude dimension and a vertical coordinate, optionally also a time
        dimension.

        The subset is extracted for a test coordinate (latitude, longitude)
        that may be specified optinally via :attr:`coord_info`.

        Parameters
        ----------
        **coord_info
            optional test coordinate specifications for other than vertical
            dimension. For all dimensions that are not specified explicitely,
            the first available coordinate in :attr:`data_obj` is used.
        """

        d = self.data_obj
        if not d.has_latlon_dims:
            raise DataDimensionError('Gridded data object needs both latitude '
                                     'and longitude dimensions')
        try:
            d.check_dimcoords_tseries()
        except DataDimensionError:
            d.reorder_dimensions_tseries()
        test_coord = {}
        for dim_coord in d.dimcoord_names[:-1]:
            if dim_coord in coord_info:
                test_coord[dim_coord] = coord_info[dim_coord]
            else:
                test_coord[dim_coord] = d[dim_coord].points[0]
        subset = d.sel(**test_coord)
        if not subset.ndim == 1:
            raise DataDimensionError('Something went wrong with extraction of '
                                     '1D subset at coordinate {}. Resulting '
                                     'data object has {} dimensions instead'
                                     .format(test_coord, subset.ndim))
        self._subset1d = subset
        return subset

    def search_aux_coords(self, coord_list):
        """Search and assign coordinates provided by input list

        All coordinates that are found are assigned to this object and can
        be accessed via ``self[coord_name]``.

        Parameters
        ----------
        coord_list : list
            list containing AeroCom coordinate names

        Returns
        -------
        bool
            True if all coordinates can be accessed, else False

        Raises
        ------
        CoordinateNameError
            if one of the input coordinate names is not supported by pyaerocom.
            See coords.ini file of pyaerocom for available coordinates.
        """
        all_ok = True
        if isinstance(coord_list, str):
            coord_list = [coord_list]
        d = self.data_obj
        for coord in coord_list:
            if coord in self:
                pass
            try:
                coord_info = const.COORDINFO[coord]
            except VariableDefinitionError:
                raise CoordinateNameError('Coordinate {} is not supported by '
                                          'pyaerocom.'.format(coord))
            # 1. check if coordinate is assigned in data object directly
            d._update_coord_info()
            if coord in d._coord_var_names:
                self[coord] = d[coord]
            else:
                try:
                    self[coord] = d.search_other(coord)
                    print('Adding coord {}'.format(coord))
                except Exception:
                    all_ok = False
        return all_ok

    @property
    def reader(self):
        """Instance of :class:`ReadGridded`"""
        return self.data_obj.reader

    def _check_vars_in_data_obj(self):
        for var in self.ADD_FILE_VARS:
            try:
                self._check_var_in_data_obj(var_name=var)
                return var
            except Exception:
                pass
        raise VariableNotFoundError()

    # ToDo: check alias names
    def _check_var_in_data_obj(self, var_name):

        c = VerticalCoordinate(var_name)

        if c.var_name in self.data_obj:
            self[var_name] = self.data_obj[var_name]
            self._verify_altitude_access()
        elif var_name in c.STANDARD_NAMES:
            std_name = c.STANDARD_NAMES[var_name]
            if std_name in self.data_obj:
                self[var_name] = self.data_obj[std_name]

    def check_altitude_access(self, **coord_info):
        """Checks if altitude levels can be accessed

        Parameters
        ----------
        **coord_info
            test coordinate specifications for extraction of 1D data object.
            Passed to :func:`extract_1D_subset_from_data`.

        Returns
        -------
        bool
            True, if altitude access is provided, else False
        """
        # 1. check if altitude or pressure field is available in data object as
        # variable
        for coord in self.coord_list:
            if coord in self._checked_and_failed:
                continue
            elif self._check_altitude_access_helper(coord, **coord_info):
                return True
            self._checked_and_failed.append(coord)
        return False

    def _check_altitude_access_helper(self, coord_name, **coord_info):

        cstd_name = const.COORDINFO[coord_name].standard_name

        if not self.search_aux_coords(coord_name):
            if isinstance(cstd_name, str):
                if not self.search_aux_coords(cstd_name):
                    return False

        coord = VerticalCoordinate(cstd_name)
        if coord.conversion_supported:
            if self.search_aux_coords(coord.conversion_requires):
                if self._verify_altitude_access(coord):
                    return True
        return False

    def _verify_altitude_access(self, coord, **coord_info):
        """Verify access of altitude data

        Parameters
        ----------
        subset : GriddedData
            1-dimensional subset of input data object
        coord : VerticalCoordinate
            instance of vertical coordinate that is used to specify requirements
            for altitude computation

        Returns
        -------
        bool
            True, if altitude access was sueccessful, else False
        """
        subset = self._subset1d
        if subset is None:
            subset = self.extract_1D_subset_from_data(**coord_info)

        subset._update_coord_info()
        cstd_name = coord.standard_name
        if not subset[cstd_name].ndim == 1:
            raise DataDimensionError('Unexpected error: dimension of variable '
                                     '{} should be 1'.format(cstd_name))
        raise NotImplementedError

    def get_altitude(self, latitude, longitude):
        raise NotImplementedError

# =============================================================================
#     ### ToDo: CHECK IF THE FOLLOWING METHODS WILL BE REQUIRED
#     def find_and_import_auxvars(self, reader=None):
#         """Find and read auxiliary variables for altitude access
#
#         This directory (if available) is used to check if potential files with
#         altitude information are available
#
#         Parameters
#         ----------
#         reader : ReadGridded
#             instance of reader class, if None, then the reader is accessed from
#             :attr:`data_obj` (instance of :class:`GriddedData`)
#         """
#         if reader is None:
#             reader = self.data_obj.reader
#         for add_var in self.ADD_FILE_VARS:
#             try:
#                 self.check_and_import_add_var(add_var)
#             except (VariableNotFoundError, AltitudeAccessError) as e:
#                 self.logger.warning(repr(e))
#         raise CoordinateError('Could not find required additional coordinate '
#                               'information for the computation of altitude '
#                               'levels')
#
#     def check_and_import_add_var(self, add_var):
#
#         add_var_data = self._find_and_read_add_var(add_var)
#         _req = []
#         if add_var in self.ADD_FILE_REQ:
#             for aux_var in self.ADD_FILE_REQ[add_var]:
#                 _req[aux_var] = self._find_and_read_add_var(aux_var)
#         _opt = []
#         if add_var in self.ADD_FILE_OPT:
#             for aux_opt in self.ADD_FILE_OPT[add_var]:
#                 try:
#                     _opt[aux_opt] = self._find_and_read_add_var(aux_opt)
#                 except Exception:
#                     pass
#
#         self._check_altitude_access(add_var_data,
#                                     add_var_data_req=_req,
#                                     add_var_data_opt=_opt)
#         self[add_var] = add_var_data
#         for item in _req:
#             self[item.var_name] = item
#         for item in _opt:
#             self[item.var_name] = item
#
#     def _find_and_read_add_var(self, add_var_name):
#         """Search an additionally required variable in external file
#
#         Parameter
#         --------
#         add_var_name : str
#             Variable name in AEROCOM convention
#
#         Returns
#         -------
#         GriddedData
#             additional data file
#         """
#         _search = [add_var_name]
#         _search.extend(const.VARS[add_var_name].aliases)
#         r = self.data_obj.reader
#         d = self.data_obj
#         for _var in _search:
#             if _var in r.vars:
#                 return r.read_var(_var, start=d.start, stop=d.stop,
#                                   ts_type=d.ts_type, flex_ts_type=False)
#         raise VariableNotFoundError('Auxiliary variable {} could not be found'
#                                     .format(add_var_name))
#
#     def _check_coord_conversion(self, add_var_data, add_var_data_req,
#                                 add_var_data_opt):
#
#         coord = VerticalCoordinate(add_var_data.standard_name)
#
#     def _check_altitude_access(self, add_var_data, add_var_data_req=None,
#                                add_var_data_opt=None):
#         """Checks if altitude can be computed based on input fields
#
#         See :func:`check_and_import_add_var` for details on usage.
#
#         """
#         var = add_var_data.var_name
#         if var == 'z':
#             if not add_var_data.shape == self.data_obj.shape:
#                 raise NotImplementedError('Altitude field must have the same '
#                                           'shape as input data object')
#             unit_std = get_standard_unit(add_var_data.var_name)
#             if not add_var_data.unit == unit_std:
#                 add_var_data.convert_unit(unit_std)
#             if add_var_data.standard_name == 'altitude':
#                 return True
#             elif add_var_data.standard_name == 'geopotential_height':
#                 self._check_coord_conversion(add_var_data, add_var_data_req,
#                                              add_var_data_opt)
# =============================================================================

if __name__ == '__main__':
    import pyaerocom as pya

# =============================================================================
#     # has pressure field directly
#     reader0 = pya.io.ReadGridded('GISS-MATRIX_GLOFIR0p5')
#
#     d0 = reader0.read_var('ec550aer',
#                           vert_which='ModelLevel',
#                           start=2008)
#
#     d0.altitude_access.check_altitude_access()
# =============================================================================
    #d0.reorder_dimensions_tseries()
    #d0 = d0[0]

    #subset = d0.sel(latitude=30, longitude=15)

    #p = d0['air_pressure']

    reader1 = pya.io.ReadGridded('ECHAM6-SALSA_CTRL2016-PD')

    d1 = reader1.read_var('od550aer3d',
                          vert_which='ModelLevel',
                          start=2010)

    aac1 = d1.altitude_access
    aac1.check_altitude_access()
# =============================================================================
#     reader2 = pya.io.ReadGridded('OsloCTM3v1.01')
#
#     d2 = reader2.read_var('ec550aer',
#                           vert_which='ModelLevel',
#                           start=2010)
#     d2['lon']
#     aac2 = d2.altitude_access
#
#     subset2 = aac2.extract_1D_subset_from_data()
#
#
#     #print(subset1)
#     print(subset2)
# =============================================================================

# =============================================================================
#     import numpy as np
#     import pyaerocom as pya
#
#     pya.change_verbosity('warning')
#     c = VerticalCoordinate('pressure')
#
#     arr = np.linspace(1013*100, 100*100, 100)
#     print(arr)
#
#     print(c.calc_altitude(p=arr))
#
#
#     r = pya.io.ReadGridded('CAM5.3-Oslo_AP3-CTRL2016-PD')
#     print(r)
#     data = r.read_var('ec5503Daer')
#
#     acc = AltitudeAccess(data)
#     acc.find_and_import_auxvars()
#
#
# =============================================================================
